#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import os
import logging
import uuid as sys_uuid
import typing as tp

import bazooka
from restalchemy.storage import base as ra_storage

from gcl_sdk.clients.http import base
from gcl_sdk.agents.universal.drivers import direct
from gcl_sdk.agents.universal.storage import fs

from gcl_sdk.agents.universal.clients.backend import core as core_rest_back
from gcl_sdk.agents.universal.clients.backend import db as db_back
from gcl_sdk.agents.universal import constants as c

LOG = logging.getLogger(__name__)


CORE_TARGET_FIELDS_FILENAME = "core_target_fields.json"


# DEPRECATED(akremenetsky): Use RestCoreCapabilityDriver instead.
class CoreCapabilityDriver(direct.DirectAgentDriver):
    """Core capability driver for interacting with Genesis Core."""

    def __init__(
        self,
        username: str,
        password: str,
        project_id: str,
        user_api_base_url: str,
        agent_work_dir: str = c.WORK_DIR,
        **collection_map,
    ):
        http = bazooka.Client()
        auth = base.CoreIamAuthenticator(
            user_api_base_url, username, password, http_client=http
        )
        self._collection_map = {
            k: v.strip() for k, v in collection_map.items()
        }

        rest_client = base.CollectionBaseClient(
            http_client=http, base_url=user_api_base_url, auth=auth
        )

        storage_path = os.path.join(
            agent_work_dir, CORE_TARGET_FIELDS_FILENAME
        )

        storage = fs.TargetFieldsFileStorage(storage_path)
        rest_client = core_rest_back.GCRestApiBackendClient(
            rest_client, collection_map, project_id
        )

        super().__init__(storage=storage, client=rest_client)

    def get_capabilities(self) -> list[str]:
        """Returns a list of capabilities supported by the driver."""
        return list(self._collection_map.keys())


class RestCoreCapabilityDriver(CoreCapabilityDriver):
    """Core capability driver for interacting with GC using REST API."""

    pass


class DatabaseCapabilityDriver(direct.DirectAgentDriver):
    """Database capability driver for interacting with GC using database."""

    def __init__(
        self,
        models: tp.Collection[
            tuple[tp.Type[ra_storage.AbstractStorableMixin], str]
        ],
        project_id: sys_uuid.UUID,
        target_fields_storage_path: str,
    ):

        model_specs = db_back.ModelSpec.from_collection(models, project_id)
        client = db_back.DatabaseBackendClient(model_specs)

        storage = fs.TargetFieldsFileStorage(target_fields_storage_path)

        self._kinds = {m.kind for m in model_specs}

        super().__init__(storage=storage, client=client)

    def get_capabilities(self) -> list[str]:
        """Returns a list of capabilities supported by the driver."""
        return list(self._kinds)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import abc
import typing as tp
import uuid as sys_uuid
import http as httplib
from urllib.parse import urljoin

import bazooka
from bazooka import exceptions as bazooka_exc
from restalchemy.dm import models


class AbstractAuthenticator(abc.ABC):

    @abc.abstractmethod
    def authenticate(self) -> None:
        """Authenticate the client."""

    @abc.abstractmethod
    def get_auth_header(self) -> dict[str, str]:
        """Get the authentication header."""


class CoreIamAuthenticator(AbstractAuthenticator):
    DEFAULT_CLIENT_UUID = sys_uuid.UUID("00000000-0000-0000-0000-000000000000")

    def __init__(
        self,
        base_url: str,
        username: str,
        password: str,
        client_id: str = "GenesisCoreClientId",
        client_secret: str = "GenesisCoreClientSecret",
        client_uuid: sys_uuid.UUID = DEFAULT_CLIENT_UUID,
        scope: str | None = None,
        ttl: int = 86400,  # 1 day
        http_client: bazooka.Client | None = None,
    ):
        self._http_client = http_client or bazooka.Client()

        client_uuid = str(client_uuid)
        self._url = (
            f"{base_url}/v1/iam/clients/{client_uuid}/actions/get_token/invoke"
        )

        self._headers = {"Content-Type": "application/x-www-form-urlencoded"}

        self._data = {
            "grant_type": "password",
            "username": username,
            "password": password,
            "client_id": client_id,
            "client_secret": client_secret,
            "scope": scope or self.empty_scope(),
            "ttl": str(ttl),
        }

        self._access_token = None
        self._refresh_token = None

    def authenticate(self) -> None:
        response = self._http_client.post(
            self._url, headers=self._headers, data=self._data
        )
        data = response.json()
        self._access_token = data["access_token"]
        self._refresh_token = data["refresh_token"]

    def get_auth_header(self) -> dict[str, str]:
        if not self._access_token:
            self.authenticate()
        return {"Authorization": f"Bearer {self._access_token}"}

    @classmethod
    def empty_scope(cls) -> str:
        return ""

    @classmethod
    def project_scope(cls, project_id: sys_uuid.UUID) -> str:
        return f"project:{project_id}"


class CollectionBaseClient:
    ACTIONS_KEY = "actions"
    INVOKE_KEY = "invoke"

    def __init__(
        self,
        base_url: str,
        http_client: bazooka.Client | None = None,
        auth: AbstractAuthenticator | None = None,
    ) -> None:
        self._http_client = http_client or bazooka.Client()
        self._base_url = base_url
        self._auth = auth

    def _collection_url(self, collection: str):
        if not self._base_url.endswith("/") and not collection.startswith("/"):
            return self._base_url + "/" + collection
        return self._base_url + collection

    def _request(
        self,
        method: httplib.HTTPMethod,
        url: str,
        data: dict[str : tp.Any] | None = None,
        json: dict[str : tp.Any] | None = None,
        params: dict[str : tp.Any] | None = None,
        headers: dict[str:str] | None = None,
    ):
        if method == httplib.HTTPMethod.POST:
            requester = self._http_client.post
        elif method == httplib.HTTPMethod.GET:
            requester = self._http_client.get
        elif method == httplib.HTTPMethod.PUT:
            requester = self._http_client.put
        elif method == httplib.HTTPMethod.DELETE:
            requester = self._http_client.delete
        else:
            raise ValueError(f"Method {method} is not supported")

        if self._auth is None:
            return requester(
                url, data=data, json=json, params=params, headers=headers
            )

        headers = headers or {}
        headers.update(self._auth.get_auth_header())

        try:
            return requester(
                url, data=data, json=json, params=params, headers=headers
            )
        except bazooka_exc.UnauthorizedError:
            # Perhaps we need to re-authenticate
            self._auth.authenticate()
            headers.update(self._auth.get_auth_header())
            return requester(
                url, data=data, json=json, params=params, headers=headers
            )

    def resource_url(self, collection: str, uuid: sys_uuid.UUID):
        return urljoin(self._collection_url(collection), str(uuid))

    def get(self, collection: str, uuid: sys_uuid.UUID) -> dict[str : tp.Any]:
        url = self.resource_url(collection, uuid)
        resp = self._request(httplib.HTTPMethod.GET, url)
        return resp.json()

    def filter(
        self, collection: str, **filters: tp.Dict[str, tp.Any]
    ) -> list[dict[str : tp.Any]]:
        resp = self._request(
            httplib.HTTPMethod.GET,
            self._collection_url(collection),
            params=filters,
        )
        return resp.json()

    def create(
        self, collection: str, data: dict[str : tp.Any]
    ) -> dict[str : tp.Any]:
        resp = self._request(
            httplib.HTTPMethod.POST,
            self._collection_url(collection),
            json=data,
        )
        return resp.json()

    def update(
        self,
        collection: str,
        uuid: sys_uuid.UUID,
        **params: tp.Dict[str, tp.Any],
    ) -> dict[str : tp.Any]:
        url = self.resource_url(collection, uuid)
        resp = self._request(httplib.HTTPMethod.PUT, url, json=params)
        return resp.json()

    def delete(self, collection: str, uuid: sys_uuid.UUID) -> None:
        url = self.resource_url(collection, uuid)
        self._request(httplib.HTTPMethod.DELETE, url)

    def do_action(
        self,
        collection: str,
        name: str,
        uuid: sys_uuid.UUID,
        invoke: bool = False,
        **kwargs,
    ) -> dict[str : tp.Any] | None:
        url = self.resource_url(collection, uuid) + "/"
        action_url = urljoin(urljoin(url, self.ACTIONS_KEY) + "/", name)

        if invoke:
            action_url = urljoin(action_url + "/", self.INVOKE_KEY)
            resp = self._request(
                httplib.HTTPMethod.POST, action_url, json=kwargs
            )
        else:
            resp = self._request(
                httplib.HTTPMethod.GET, action_url, params=kwargs
            )

        # Try to convert response to json
        resp.raise_for_status()
        try:
            return resp.json()
        except bazooka_exc.BaseHTTPException:
            return None


class CollectionBaseModelClient(CollectionBaseClient):
    __model__: tp.Type[models.SimpleViewMixin] = None
    __resource_client__: tp.Type["ResourceBaseModelClient"] = None
    __parent__: str | None = None

    def __init__(
        self,
        base_url: str,
        collection_path: str,
        http_client: bazooka.Client | None = None,
        auth: AbstractAuthenticator | None = None,
    ) -> None:
        super().__init__(base_url, http_client, auth)
        self._collection_path = collection_path

    def get_collection(self) -> str:
        return self._collection_path

    def __call__(
        self, resource_uuid: sys_uuid.UUID
    ) -> "ResourceBaseModelClient":
        if self.__resource_client__ is None:
            raise ValueError("Resource client is not defined")
        return self.__resource_client__(self, resource_uuid, self._http_client)

    def get(self, uuid: sys_uuid.UUID) -> models.SimpleViewMixin:
        return self.__model__.restore_from_simple_view(
            **super().get(self.get_collection(), uuid)
        )

    def filter(
        self, **filters: tp.Dict[str, tp.Any]
    ) -> list[models.SimpleViewMixin]:
        return [
            self.__model__.restore_from_simple_view(**o)
            for o in super().filter(self.get_collection(), **filters)
        ]

    def create(self, object: models.SimpleViewMixin) -> models.SimpleViewMixin:
        skip = tuple() if self.__parent__ is None else (self.__parent__,)
        data = object.dump_to_simple_view(skip=skip)
        return self.__model__.restore_from_simple_view(
            **super().create(self.get_collection(), data)
        )

    def update(
        self, uuid: sys_uuid.UUID, **params: tp.Dict[str, tp.Any]
    ) -> models.SimpleViewMixin:
        return self.__model__.restore_from_simple_view(
            **super().update(self.get_collection(), uuid, **params)
        )

    def delete(self, uuid: sys_uuid.UUID) -> None:
        super().delete(self.get_collection(), uuid)

    def do_action(
        self, name: str, uuid: sys_uuid.UUID, invoke: bool = False, **kwargs
    ) -> dict[str : tp.Any] | None:
        return super().do_action(
            self.get_collection(), name, uuid, invoke, **kwargs
        )


class StaticCollectionBaseModelClient(CollectionBaseModelClient):
    __collection_path__: str | None = None

    def __init__(
        self,
        base_url: str,
        http_client: bazooka.Client | None = None,
        auth: AbstractAuthenticator | None = None,
    ) -> None:
        super().__init__(base_url, self.__collection_path__, http_client, auth)


class ResourceBaseModelClient:
    ACTIONS_KEY = "actions"
    INVOKE_KEY = "invoke"

    __model__: tp.Type[models.SimpleViewMixin] = None

    def __init__(
        self,
        collection: StaticCollectionBaseModelClient,
        resource_uuid: sys_uuid.UUID,
        http_client: bazooka.Client | None = None,
    ):
        self._http_client = http_client or bazooka.Client()
        self._collection = collection
        self._resource_uuid = resource_uuid

    def resource_url(self):
        return self._collection.resource_url(
            self._collection.__collection_path__, self._resource_uuid
        )

    def get(self) -> models.SimpleViewMixin:
        url = self.resource_url()
        resp = self._http_client.get(url)
        return self.__model__.restore_from_simple_view(**resp.json())

    def update(self, **params: tp.Dict[str, tp.Any]) -> models.SimpleViewMixin:
        url = self.resource_url()
        resp = self._http_client.put(url, json=params)
        return self.__model__.restore_from_simple_view(**resp.json())

    def delete(self) -> None:
        url = self.resource_url()
        self._http_client.delete(url)

    def do_action(
        self, name: str, invoke: bool = False, **kwargs
    ) -> tp.Dict[str, tp.Any] | None:
        url = self.resource_url() + "/"
        action_url = urljoin(urljoin(url, self.ACTIONS_KEY) + "/", name)

        if invoke:
            action_url = urljoin(action_url + "/", self.INVOKE_KEY)
            resp = self._http_client.post(action_url, json=kwargs)
        else:
            resp = self._http_client.get(action_url, params=kwargs)

        # Try to convert response to json
        resp.raise_for_status()
        try:
            return resp.json()
        except bazooka_exc.BaseHTTPException:
            return None



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import os
import json
import xxhash
import typing as tp
import uuid as sys_uuid

from gcl_sdk.agents.universal import constants as c


def system_uuid(
    system_uuid_path: str = "/sys/class/dmi/id/product_uuid",
) -> sys_uuid.UUID:
    """Return system uuid"""
    with open(system_uuid_path) as f:
        return sys_uuid.UUID(f.read().strip())


def node_uuid(
    node_path: str = c.NODE_UUID_PATH, use_machine_if_absent: bool = True
) -> sys_uuid.UUID:
    """Return node uuid"""
    if os.path.exists(node_path):
        with open(node_path) as f:
            return sys_uuid.UUID(f.read().strip())

    if use_machine_if_absent:
        return system_uuid()

    raise FileNotFoundError(f"The node-id location {node_path} not found")


def calculate_hash(
    value: dict, hash_method: tp.Callable[[str], str] = xxhash.xxh3_64
) -> str:
    m = hash_method()
    m.update(
        json.dumps(value, separators=(",", ":"), sort_keys=True).encode(
            "utf-8"
        )
    )
    return m.hexdigest()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.api import routes

from gcl_sdk.agents.universal.orch_api import controllers


class UniversalAgentsGetPayloadAction(routes.Action):
    """Handler for /v1/agents/<uuid>/actions/get_payload endpoint"""

    __controller__ = controllers.UniversalAgentsController


class UniversalAgentsRoute(routes.Route):
    """Handler for /v1/agents/ endpoint"""

    __controller__ = controllers.UniversalAgentsController

    get_payload = routes.action(UniversalAgentsGetPayloadAction)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import itertools
import uuid as sys_uuid

import pytest

from restalchemy.dm import models as ra_models
from restalchemy.dm import properties
from restalchemy.dm import types as ra_types
from restalchemy.api import applications
from restalchemy.api import middlewares
from restalchemy.api import controllers
from restalchemy.api import routes
from restalchemy.api.middlewares import contexts as context_mw
from restalchemy.api.middlewares import logging as logging_mw
from restalchemy.api.middlewares import errors as errors_mw
from restalchemy.openapi import structures as openapi_structures
from restalchemy.openapi import engines as openapi_engines

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal import constants as c
from gcl_sdk.agents.universal.orch_api import routes as orch_routes
from gcl_sdk.agents.universal.status_api import routes as status_routes
from gcl_sdk.audit.api import routes as audit_routes

FIRST_MIGRATION = "0000-init-events-table-2cfd220e.py"


class FooResource(ra_models.ModelWithUUID, models.ResourceMixin):
    name = properties.property(
        ra_types.String(max_length=64), default="foo-name"
    )
    project_id = properties.property(
        ra_types.UUID(), default=lambda: sys_uuid.uuid4()
    )


class FooTargetResource(ra_models.ModelWithUUID, models.TargetResourceMixin):
    name = properties.property(
        ra_types.String(max_length=64), default="foo-name"
    )
    project_id = properties.property(
        ra_types.UUID(), default=lambda: sys_uuid.uuid4()
    )


class DummyInstance(
    ra_models.ModelWithUUID,
    ra_models.ModelWithTimestamp,
    models.InstanceMixin,
):
    NAMESPACE = sys_uuid.UUID("14cf9662-6297-4ce8-92aa-9568a75f352e")

    __dummy_storage__ = {
        "new": [],
        "updated": [],
        "deleted": [],
        "existing": [],
    }

    name = properties.property(
        ra_types.String(max_length=64), default="foo-name"
    )
    project_id = properties.property(
        ra_types.UUID(), default=lambda: sys_uuid.uuid4()
    )

    def save(self, **kwargs):
        pass

    def delete(self, **kwargs):
        pass

    def update(self, **kwargs):
        pass

    @classmethod
    def get_resource_kind(cls):
        return "foo"

    @classmethod
    def _has_model_derivatives(cls) -> bool:
        return False

    @classmethod
    def get_new_instances(
        cls, limit: int = c.DEF_SQL_LIMIT
    ) -> list["DummyInstance"]:
        return cls.__dummy_storage__["new"]

    @classmethod
    def get_updated_instances(
        cls, limit: int = c.DEF_SQL_LIMIT
    ) -> list["DummyInstance"]:
        return cls.__dummy_storage__["updated"]

    @classmethod
    def get_deleted_instances(
        cls, limit: int = c.DEF_SQL_LIMIT
    ) -> list["models.TargetResource"]:
        return cls.__dummy_storage__["deleted"]


class DummyDerivative(
    ra_models.ModelWithUUID,
    models.TargetResourceKindAwareMixin,
):
    NAMESPACE = sys_uuid.UUID("5bf3b9e4-8975-44d9-ade5-e7e1abcf408a")

    __dummy_storage__ = {
        "new": [],
        "updated": [],
        "deleted": [],
        "existing": [],
    }

    name = properties.property(
        ra_types.String(max_length=64), default="foo-name"
    )
    project_id = properties.property(
        ra_types.UUID(), default=lambda: sys_uuid.uuid4()
    )

    @classmethod
    def get_resource_kind(cls) -> str:
        return "foo-derivative"


class DummyInstanceWithDerivatives(
    DummyInstance, models.InstanceWithDerivativesMixin
):
    __derivative_model_map__ = {
        "foo-derivative": DummyDerivative,
    }

    @classmethod
    def _has_model_derivatives(cls) -> bool:
        return True

    def to_derivative(self) -> DummyDerivative:
        return DummyDerivative(
            uuid=sys_uuid.uuid5(self.uuid, self.name),
            name=self.name,
            project_id=self.project_id,
        )


class DummySubInstanceWithDerivatives(
    DummyInstance, models.InstanceWithDerivativesMixin
):
    __master_model__ = DummyInstanceWithDerivatives
    __derivative_model_map__ = {
        "foo-derivative": DummyDerivative,
    }

    @classmethod
    def _has_model_derivatives(cls) -> bool:
        return True

    @classmethod
    def get_resource_kind(cls) -> str:
        return "sub-foo"

    def to_derivative(self) -> DummyDerivative:
        return DummyDerivative(
            uuid=sys_uuid.uuid5(self.uuid, self.name),
            name=self.name,
            project_id=self.project_id,
        )


def get_openapi_engine():
    openapi_engine = openapi_engines.OpenApiEngine(
        info=openapi_structures.OpenApiInfo(
            title="Test API",
            version="v1",
            description="OpenAPI - Test API",
        ),
        paths=openapi_structures.OpenApiPaths(),
        components=openapi_structures.OpenApiComponents(),
    )
    return openapi_engine


@pytest.fixture(scope="module")
def orch_api_wsgi_app():
    class OrchApiApp(routes.RootRoute):
        pass

    class ApiEndpointController(controllers.RoutesListController):
        __TARGET_PATH__ = "/v1/"

    class ApiEndpointRoute(routes.Route):
        __controller__ = ApiEndpointController
        __allow_methods__ = [routes.FILTER]

        agents = routes.route(orch_routes.UniversalAgentsRoute)

    setattr(
        OrchApiApp,
        "v1",
        routes.route(ApiEndpointRoute),
    )

    return middlewares.attach_middlewares(
        applications.OpenApiApplication(
            route_class=OrchApiApp,
            openapi_engine=get_openapi_engine(),
        ),
        [
            context_mw.ContextMiddleware,
            errors_mw.ErrorsHandlerMiddleware,
            logging_mw.LoggingMiddleware,
        ],
    )


@pytest.fixture(scope="module")
def status_api_wsgi_app():
    class StatusApiApp(routes.RootRoute):
        pass

    class ApiEndpointController(controllers.RoutesListController):
        __TARGET_PATH__ = "/v1/"

    class ApiEndpointRoute(routes.Route):
        __controller__ = ApiEndpointController
        __allow_methods__ = [routes.FILTER]

        agents = routes.route(status_routes.UniversalAgentsRoute)
        kind = routes.route(status_routes.KindRoute)

    setattr(
        StatusApiApp,
        "v1",
        routes.route(ApiEndpointRoute),
    )

    return middlewares.attach_middlewares(
        applications.OpenApiApplication(
            route_class=StatusApiApp,
            openapi_engine=get_openapi_engine(),
        ),
        [
            context_mw.ContextMiddleware,
            errors_mw.ErrorsHandlerMiddleware,
            logging_mw.LoggingMiddleware,
        ],
    )


@pytest.fixture(scope="module")
def audit_api_wsgi_app():
    class AuditApiApp(routes.RootRoute):
        pass

    class ApiEndpointController(controllers.RoutesListController):
        __TARGET_PATH__ = "/v1/"

    class ApiEndpointRoute(routes.Route):
        __controller__ = ApiEndpointController
        __allow_methods__ = [routes.FILTER]

        audit = routes.route(audit_routes.AuditRoute)

    setattr(
        AuditApiApp,
        "v1",
        routes.route(ApiEndpointRoute),
    )

    return middlewares.attach_middlewares(
        applications.OpenApiApplication(
            route_class=AuditApiApp,
            openapi_engine=get_openapi_engine(),
        ),
        [
            context_mw.ContextMiddleware,
            errors_mw.ErrorsHandlerMiddleware,
            logging_mw.LoggingMiddleware,
        ],
    )


@pytest.fixture
def dummy_instance_factory():

    def factory(
        spec: dict | None = None, project_id: sys_uuid.UUID | None = None
    ) -> list[DummyInstance]:
        storage = {
            "new": [],
            "updated": [],
            "deleted": [],
            "existing": [],
        }
        project_id = project_id or sys_uuid.uuid4()

        if spec is None:
            spec = {"new": 1}

        for mode, count in spec.items():
            for i in range(count):
                uuid = sys_uuid.uuid5(DummyInstance.NAMESPACE, f"{mode}-{i}")
                name = f"inst-{mode}-{i}"

                instance = DummyInstance(
                    uuid=uuid, name=name, project_id=project_id
                )
                if mode == "deleted":
                    instance = instance.to_ua_resource()
                elif mode == "existing":
                    resource = instance.to_ua_resource()
                    resource.save()
                elif mode == "updated":
                    resource = instance.to_ua_resource()
                    instance.name = f"{name}-updated"
                    resource.save()

                storage[mode].append(instance)

        DummyInstance.__dummy_storage__ = storage

        return list(itertools.chain.from_iterable(storage.values()))

    yield factory

    DummyInstance.__dummy_storage__ = {
        "new": [],
        "updated": [],
        "deleted": [],
        "existing": [],
    }


@pytest.fixture
def dummy_instance_with_derivatives_factory():

    def factory(
        spec: dict | None = None,
        project_id: sys_uuid.UUID | None = None,
        model: type[DummyInstance] = DummyInstanceWithDerivatives,
    ) -> list[DummyInstance]:
        storage = {
            "new": [],
            "updated": [],
            "deleted": [],
            "existing": [],
        }
        project_id = project_id or sys_uuid.uuid4()

        if spec is None:
            spec = {"new": 1}

        for mode, count in spec.items():
            for i in range(count):
                uuid = sys_uuid.uuid5(model.NAMESPACE, f"{mode}-{i}")
                name = f"inst-{mode}-{i}"

                instance = model(uuid=uuid, name=name, project_id=project_id)
                if master_model := getattr(model, "__master_model__", None):
                    master_instance = master_model(
                        uuid=uuid, name=name, project_id=project_id
                    )
                    master_resource = master_instance.to_ua_resource()
                    master_instance.save()
                    master_resource.save()
                else:
                    master_instance = None

                if mode == "deleted":
                    instance = instance.to_ua_resource()
                elif mode == "existing":
                    derivative = instance.to_derivative()
                    resource = instance.to_ua_resource(
                        master=(
                            master_instance.uuid if master_instance else None
                        ),
                        master_hash=(
                            master_resource.hash if master_instance else ""
                        ),
                        master_full_hash=(
                            master_resource.full_hash
                            if master_instance
                            else ""
                        ),
                    )
                    derivative_resource = derivative.to_ua_resource(
                        master=resource.uuid,
                        master_hash=resource.hash,
                        master_full_hash=resource.full_hash,
                    )
                    resource.save()
                    derivative_resource.save()

                elif mode == "updated":
                    derivative = instance.to_derivative()
                    resource = instance.to_ua_resource()
                    derivative_resource = derivative.to_ua_resource(
                        master=resource.uuid,
                        master_hash=resource.hash,
                        master_full_hash=resource.full_hash,
                    )
                    instance.name = f"{name}-updated"
                    resource.save()
                    derivative_resource.save()

                storage[mode].append(instance)

        model.__dummy_storage__ = storage

        return list(itertools.chain.from_iterable(storage.values()))

    yield factory

    DummyInstanceWithDerivatives.__dummy_storage__ = {
        "new": [],
        "updated": [],
        "deleted": [],
        "existing": [],
    }

    DummySubInstanceWithDerivatives.__dummy_storage__ = {
        "new": [],
        "updated": [],
        "deleted": [],
        "existing": [],
    }



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import sys

from oslo_config import cfg

from gcl_sdk.common import config
from gcl_sdk.common import log as infra_log
from restalchemy.storage.sql import engines
from restalchemy.common import config_opts as ra_config_opts
from gcl_sdk.agents.universal.services import scheduler


DOMAIN = "universal_agent_scheduler"

svc_opts = [
    cfg.ListOpt(
        "capabilities",
        default=None,
        help="List of capabilities to be scheduled",
    ),
]


CONF = cfg.CONF
CONF.register_cli_opts(svc_opts, DOMAIN)
ra_config_opts.register_posgresql_db_opts(CONF)


def main():
    # Parse config
    config.parse(sys.argv[1:])

    # Configure logging
    infra_log.configure()
    log = logging.getLogger(__name__)

    capabilities = CONF[DOMAIN].capabilities or []
    log.info("Capabilities: %s", capabilities)

    engines.engine_factory.configure_postgresql_factory(CONF)

    service = scheduler.UniversalAgentSchedulerService(
        iter_min_period=3, capabilities=capabilities
    )

    service.start()

    log.info("Bye!!!")


if __name__ == "__main__":
    main()



from __future__ import annotations

import uuid as sys_uuid
from urllib.parse import urljoin

import pytest
import requests
from gcl_iam.drivers import DummyDriver
from gcl_iam.enforcers import Enforcer
from mock import mock
from oslo_config import cfg
from restalchemy.common import contexts
from restalchemy.common.contexts import Context

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.audit.dm.models import AuditLogSQLStorableMixin, AuditRecord
from gcl_sdk.tests.functional import conftest
from gcl_sdk.tests.functional import utils as test_utils

CONF = cfg.CONF


class UniversalAgentAuditMixin(
    AuditLogSQLStorableMixin, models.UniversalAgent
): ...


introspection_info = lambda x: DummyDriver().get_introspection_info(None)


class DummyDriverIam(DummyDriver):
    introspection_info = introspection_info
    enforcer = Enforcer(["audit_log.audit_record.read"])


class DummyContext(Context):
    iam_context = DummyDriverIam()
    introspection_info = introspection_info


class TestAuditApi:

    @pytest.fixture(scope="class")
    def audit_api_service(self, audit_api_wsgi_app):
        class ApiRestService(test_utils.RestServiceTestCase):
            __FIRST_MIGRATION__ = conftest.FIRST_MIGRATION
            __APP__ = audit_api_wsgi_app

        rest_service = ApiRestService()
        rest_service.setup_class()
        yield rest_service
        rest_service.teardown_class()

    @pytest.fixture()
    def audit_api(self, audit_api_service: test_utils.RestServiceTestCase):
        audit_api_service.setup_method()
        yield audit_api_service
        audit_api_service.teardown_method()

    def test_no_audit(self, audit_api: test_utils.RestServiceTestCase):
        url = urljoin(audit_api.base_url, "audit/")
        contexts.get_context = mock.MagicMock(return_value=DummyContext)
        response = requests.get(url)
        assert response.json() == []
        assert response.status_code == 200

    def test_audit_get(self, audit_api: test_utils.RestServiceTestCase):
        uuid_a = sys_uuid.uuid4()
        agent_a = UniversalAgentAuditMixin(
            name="Agent A", uuid=uuid_a, node=uuid_a
        )
        agent_a.insert()
        audit = AuditRecord.objects.get_one(
            filters={"object_uuid": agent_a.uuid}
        )
        url = urljoin(audit_api.base_url, f"audit/{audit.uuid}")
        contexts.get_context = mock.MagicMock(return_value=DummyContext)
        response = requests.get(url)
        output = response.json()
        expected = {
            "action": "create",
            "object_type": "ua_agents",
            "object_uuid": str(agent_a.uuid),
            "uuid": str(audit.uuid),
        }
        assert response.status_code == 200
        for key in expected.keys():
            assert output[key] == expected[key]

    def test_audit_list(self, audit_api: test_utils.RestServiceTestCase):
        uuid_a = sys_uuid.uuid4()
        agent_a = UniversalAgentAuditMixin(
            name="Agent A", uuid=uuid_a, node=uuid_a
        )
        uuid_b = sys_uuid.uuid4()
        agent_b = UniversalAgentAuditMixin(
            name="Agent B", uuid=uuid_b, node=uuid_b
        )
        agent_a.insert()
        agent_b.insert()
        audits = AuditRecord.objects.get_all()
        url = urljoin(audit_api.base_url, "audit/")
        contexts.get_context = mock.MagicMock(return_value=DummyContext)
        response = requests.get(url)
        output = response.json()
        assert response.status_code == 200
        assert len(output) == 2
        assert {a["uuid"] for a in output} == {
            str(audits[0].uuid),
            str(audits[1].uuid),
        }



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json
import os
import uuid as sys_uuid

import pytest

from gcl_sdk.agents.universal.drivers import exceptions as driver_exc
from gcl_sdk.agents.universal.drivers import meta
from gcl_sdk.agents.universal.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types


def _make_resource(
    kind: str, uuid: sys_uuid.UUID | None = None, value: dict | None = None
) -> models.Resource:
    uuid = uuid or sys_uuid.uuid4()
    value = value or {"uuid": str(uuid), "foo": 1}
    return models.Resource.from_value(
        value, kind, target_fields=frozenset(value.keys())
    )


class DummyModel(meta.MetaDataPlaneModel):
    """A lightweight DP model for testing meta driver flows.

    It records calls to DP-like methods in a class-level log.
    """

    call_log: dict[str, list[str]] = {}

    # Custom fields that may appear in resource.value
    foo = properties.property(types.Integer(), default=0)
    raise_on_restore = properties.property(types.Boolean(), default=False)

    def get_meta_model_fields(self) -> set[str] | None:
        # Save all fields into meta file (including any service flags)
        return None

    def _log(self, action: str) -> None:
        self.call_log.setdefault(str(self.uuid), []).append(action)

    def dump_to_dp(self) -> None:  # create
        self._log("dump_to_dp")

    def restore_from_dp(self) -> None:  # get/list
        # Optional flag to simulate not-found on DP for this object
        if getattr(self, "raise_on_restore", False):
            raise driver_exc.ResourceNotFound(
                resource=models.Resource.from_value(
                    {"uuid": str(self.uuid)}, "dummy", frozenset({"uuid"})
                )
            )
        self._log("restore_from_dp")

    def delete_from_dp(self) -> None:
        self._log("delete_from_dp")

    def update_on_dp(self) -> None:
        self._log("update_on_dp")


class _Driver(meta.MetaFileStorageAgentDriver):
    __model_map__ = {"dummy": DummyModel}


class TestMetaDriver:
    def test_finalize_persists_meta_file(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        # Initially empty storage
        assert drv._storage == {}

        # Create one resource to populate meta storage
        res = _make_resource("dummy")
        drv.create(res)

        # Persist to file
        drv.finalize()

        assert os.path.exists(meta_file)
        with open(meta_file) as f:
            data = json.load(f)
        assert "dummy" in data and "resources" in data["dummy"]
        assert str(res.uuid) in data["dummy"]["resources"]

    def test_get_unsupported_kind_raises(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        res = _make_resource("other")
        with pytest.raises(TypeError):
            drv.get(res)

    def test_get_not_found_raises(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        res = _make_resource("dummy")
        with pytest.raises(driver_exc.ResourceNotFound):
            drv.get(res)

    def test_create_and_get_success(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        res = _make_resource("dummy")
        created = drv.create(res)

        assert created.uuid == res.uuid
        assert created.kind == res.kind

        fetched = drv.get(res)
        assert fetched.uuid == res.uuid
        assert fetched.kind == res.kind

        # Ensure DP methods were called
        log = DummyModel.call_log[str(res.uuid)]
        assert "dump_to_dp" in log
        assert "restore_from_dp" in log

    def test_create_already_exists_maps_exception(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        res = _make_resource("dummy")
        drv.create(res)

        with pytest.raises(driver_exc.ResourceAlreadyExists):
            drv.create(res)

    def test_list_skips_objects_not_on_dp(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        # Create two resources, one will be marked to raise on restore
        uuid1, uuid2 = sys_uuid.uuid4(), sys_uuid.uuid4()
        res1 = _make_resource(
            "dummy", uuid=uuid1, value={"uuid": str(uuid1), "foo": 1}
        )
        res2 = _make_resource(
            "dummy",
            uuid=uuid2,
            value={"uuid": str(uuid2), "foo": 2, "raise_on_restore": True},
        )

        drv.create(res1)
        drv.create(res2)

        resources = drv.list("dummy")

        # Only the one that could be restored from DP should be returned
        assert len(resources) == 1
        assert resources[0].uuid == uuid1
        assert resources[0].kind == "dummy"

    def test_update_success(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        res = _make_resource("dummy")

        drv.create(res)

        # Update with new value
        new_value = {"uuid": str(res.uuid), "foo": 42}
        new_res = models.Resource.from_value(
            new_value, "dummy", target_fields=frozenset(new_value.keys())
        )

        updated = drv.update(new_res)

        assert updated.uuid == res.uuid
        assert updated.value["foo"] == 42

        # Ensure update_on_dp called and meta was refreshed
        log = DummyModel.call_log[str(res.uuid)]
        assert "update_on_dp" in log

        # The meta storage should contain updated value
        meta_obj = drv._storage["dummy"]["resources"][str(res.uuid)]
        assert meta_obj.get("foo") == 42

    def test_update_not_found_maps_exception(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        # Not created in meta -> should raise
        res = _make_resource("dummy")
        with pytest.raises(driver_exc.ResourceNotFound):
            drv.update(res)

    def test_delete_success(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        res = _make_resource("dummy")
        drv.create(res)

        # Ensure present in meta
        assert str(res.uuid) in drv._storage["dummy"]["resources"]

        drv.delete(res)

        # Ensure DP delete was called and meta entry removed
        log = DummyModel.call_log[str(res.uuid)]
        assert "delete_from_dp" in log
        assert str(res.uuid) not in drv._storage["dummy"]["resources"]

    def test_delete_unsupported_kind_raises(self, tmp_path):
        meta_file = tmp_path / "meta.json"
        drv = _Driver(meta_file=str(meta_file))

        res = _make_resource("other")
        with pytest.raises(TypeError):
            drv.delete(res)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import logging
import typing as tp
import importlib_metadata

from restalchemy.storage.sql import migrations

LOG = logging.getLogger(__name__)

EVENT_PAYLOADS_GROUP = "gcl_sdk_event_payloads"


def load_event_payload_map() -> dict:
    event_payload_map = {
        ep.name: ep.load()
        for ep in importlib_metadata.entry_points(
            group=EVENT_PAYLOADS_GROUP,
        )
    }
    return event_payload_map


def load_from_entry_point(group: str, name: str) -> tp.Any:
    """Load class from entry points."""
    for ep in importlib_metadata.entry_points(group=group):
        if ep.name == name:
            return ep.load()

    raise RuntimeError(f"No class '{name}' found in entry points {group}")


class MigrationEngine(migrations.MigrationEngine):
    """
    Helper for apply library migration from another project.
    Used from the migration file. Example:

    from gcl_sdk.common.utils import MigrationEngine
    from gcl_sdk import migrations as sdk_migrations

    SDK_MIGRATION_FILE_NAME = "0002-init-auditlog-table-c6f740.py"

    class MigrationStep(migrations.AbstarctMigrationStep):
    ...
    def upgrade(self, session):
        migration_engine = MigrationEngine._get_migration_engine(sdk_migrations)
        migration_engine.apply_migration(SDK_MIGRATION_FILE_NAME, session)

    def downgrade(self, session)
        migration_engine = MigrationEngine._get_migration_engine(sdk_migrations)
        migration_engine.rollback_migration(SDK_MIGRATION_FILE_NAME, session)
    """

    @classmethod
    def _get_migration_engine(cls, migrations_module):
        migration_path = os.path.dirname(migrations_module.__file__)
        return cls(migrations_path=migration_path)

    def apply_migration(self, migration_name, session):
        filename = self.get_file_name(migration_name)
        self._init_migration_table(session)
        migrations = self._load_migration_controllers(session)

        migration = migrations[filename]
        if migration.is_applied():
            LOG.warning("Migration '%s' is already applied", migration.name)
        else:
            LOG.info("Applying migration '%s'", migration.name)
            migrations[filename].apply(session, migrations)

    def rollback_migration(self, migration_name, session):
        filename = self.get_file_name(migration_name)
        self._init_migration_table(session)
        migrations = self._load_migration_controllers(session)
        migration = migrations[filename]
        if not migration.is_applied():
            LOG.warning("Migration '%s' is not applied", migration.name)
        else:
            LOG.info("Rolling back migration '%s'", migration.name)
            migrations[filename].rollback(session, migrations)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import uuid as sys_uuid

from bazooka import common as bzk_common
from bazooka import client as bzk_client
from bazooka import exceptions as bzk_exceptions
from izulu import root
from oslo_config import cfg
from restalchemy.dm import filters as ra_filters
from restalchemy.storage import exceptions as ra_exceptions
import yaml

from gcl_sdk.events import constants as event_c
from gcl_sdk.events.dm import models


CONF = cfg.CONF


class EventNotFound(root.Error):
    __toggles__ = root.Toggles.DEFAULT ^ root.Toggles.FORBID_UNANNOTATED_FIELDS
    __template__ = "The Event[<{event_uuid}>] not found"


class AbstractEventClient(metaclass=abc.ABCMeta):

    def __init__(self, event_type_mapping):
        self._event_type_mapping = event_type_mapping
        super().__init__()

    @classmethod
    def build_from_config(cls, conf=None, **kwags):
        conf = conf or CONF

        config_file = conf.find_file(
            conf[event_c.DOMAIN].event_type_mapping_filepath
        )

        with open(config_file) as fp:
            event_type_mapping = yaml.safe_load(fp)

        return cls(event_type_mapping=event_type_mapping, **kwags)

    def build_user_event(self, context, user, payload):
        event_type = models.UserEvent(
            user_id=user.uuid,
            uuid=self.get_event_uuid(payload.KIND),
        )
        event = models.Event(
            event_type=event_type,
            event_data=payload,
        )

        return event

    @abc.abstractmethod
    def send_event(self, event):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_event_status(self, event_uuid):
        raise NotImplementedError()

    def get_event_uuid(self, event_type_name):
        sec = self._event_type_mapping[event_c.EVENT_TYPE_MAPPING_SECTION]
        return sys_uuid.UUID(sec[event_type_name])


class DummyEventClient(AbstractEventClient):

    def __init__(self, event_type_mapping=None, **kwargs):
        event_type_mapping = event_type_mapping or {}
        super().__init__(event_type_mapping=event_type_mapping, **kwargs)

    def send_event(self, event):
        return event

    def get_event_status(self, event_uuid):
        return models.Event.STATUS.ACTIVE.value

    def get_event_uuid(self, event_type_name):
        return sys_uuid.uuid4()


class HttpEventClient(AbstractEventClient, bzk_common.RESTClientMixIn):

    def __init__(
        self,
        event_type_mapping,
        endpoint,
        version,
        auth_token,
        project_id,
        timeout=5,
    ):
        super().__init__(event_type_mapping=event_type_mapping)
        self._endpoint = endpoint
        self._version = version
        self._auth_token = auth_token
        self._project_id = project_id
        self._client = bzk_client.Client(default_timeout=timeout)

    def _get_headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._auth_token}",
        }

    def send_event(self, event):
        exchange = {}
        if isinstance(event.event_type, models.UserEvent):
            exchange = {
                "kind": "User",
                "user_id": str(event.event_type.user_id),
            }
        body = {
            "uuid": str(event.uuid),
            "exchange": exchange,
            "event_type": (
                "/%s/event_types/%s"
                % (
                    self._version,
                    event.event_type.uuid,
                )
            ),
            "event_params": event.event_data.to_simple_dict(),
            "project_id": self._project_id,
        }

        url = self._build_collection_uri([self._version, "events"])

        self._client.post(url, json=body, headers=self._get_headers())

        return event

    def get_event_status(self, event_uuid):
        url = self._build_resource_uri([self._version, "events", event_uuid])
        try:
            result = self._client.get(url, headers=self._get_headers())
            return result.json()["status"]
        except bzk_exceptions.NotFoundError:
            raise EventNotFound(event_uuid=event_uuid)

    @classmethod
    def build_from_config(cls, conf=None, **kwargs):
        conf = conf or CONF

        endpoint = conf[event_c.DOMAIN].genesis_notification_endpoint
        version = conf[event_c.DOMAIN].genesis_api_version
        auth_token = conf[event_c.DOMAIN].genesis_api_token
        project_id = conf[event_c.DOMAIN].project_id

        return super().build_from_config(
            conf=conf,
            endpoint=endpoint,
            version=version,
            auth_token=auth_token,
            project_id=project_id,
        )


class AsyncEventClient(AbstractEventClient):

    def send_event(self, event):
        event.save()
        return event

    def get_event_status(self, event_uuid):
        try:
            return models.Event.objects.get_one(
                filters={
                    "uuid": ra_filters.EQ(event_uuid),
                }
            ).status.value
        except ra_exceptions.RecordNotFound:
            raise EventNotFound(event_uuid=event_uuid)


EVENT_CLIENT_MAPPER = {
    "async": AsyncEventClient.build_from_config,
    "dummy": DummyEventClient.build_from_config,
    "http": HttpEventClient.build_from_config,
}


def build_client(conf=None):
    conf = conf or CONF

    client_type = conf[event_c.DOMAIN].client_type

    return EVENT_CLIENT_MAPPER[client_type](conf=conf)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.dm import types_dynamic
from restalchemy.storage.sql import orm

from gcl_sdk.common import constants as sdk_c
from gcl_sdk.common import utils


class UserEvent(types_dynamic.AbstractKindModel):
    KIND = "UserEvent"

    user_id = properties.property(
        types.UUID(),
        required=True,
    )
    uuid = properties.property(
        types.UUID(),
        required=True,
    )


class AbstractEventPayload(types_dynamic.AbstractKindModel):

    @abc.abstractmethod
    def to_simple_dict(self):
        raise NotImplementedError(
            "Subclasses must implement to_simple_dict method"
        )


class DummyEventPayload(AbstractEventPayload):
    KIND = "DummyEventPayload"

    def to_simple_dict(self):
        return {}


class Event(
    models.ModelWithUUID,
    models.ModelWithTimestamp,
    orm.SQLStorableMixin,
):
    __tablename__ = "gcl_sdk_events"

    STATUS = sdk_c.EventStatus

    status = properties.property(
        types.Enum([status.value for status in STATUS]),
        default=STATUS.NEW.value,
    )
    event_type = properties.property(
        types_dynamic.KindModelSelectorType(
            types_dynamic.KindModelType(UserEvent),
        ),
        required=True,
    )
    event_data = properties.property(
        types_dynamic.KindModelSelectorType(
            *[
                types_dynamic.KindModelType(ept)
                for ept in utils.load_event_payload_map().values()
            ],
        ),
        required=True,
    )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import abc
import uuid as sys_uuid
import typing as tp

from gcl_sdk.agents.universal.dm import models


class AbstractOrchClient(abc.ABC):
    """Abstract Orchestrator client.

    An interface for clients to communicate with an orchestrator.
    """

    @abc.abstractmethod
    def agents_create(
        self, agent: models.UniversalAgent, **kwargs: tp.Any
    ) -> models.UniversalAgent:
        """Create an instance of Universal agent."""

    @abc.abstractmethod
    def agents_get_payload(
        self,
        uuid: sys_uuid.UUID,
        payload: models.Payload | None,
        **kwargs: tp.Any,
    ) -> models.Payload:
        """Get payload for of the Universal agent."""

    @abc.abstractmethod
    def resources_create(
        self, resource: models.Resource, **kwargs: tp.Any
    ) -> models.Resource:
        """Create a resource."""

    @abc.abstractmethod
    def resources_get(
        self, kind: str, uuid: sys_uuid.UUID, **kwargs: tp.Any
    ) -> models.Resource:
        """Get the resource."""

    @abc.abstractmethod
    def resources_update(
        self, kind: str, uuid: sys_uuid.UUID, **kwargs: tp.Any
    ) -> models.Resource:
        """Update the resource."""

    @abc.abstractmethod
    def resources_delete(
        self, resource: models.Resource, **kwargs: tp.Any
    ) -> None:
        """Delete the resource."""



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import abc
import typing as tp
import uuid as sys_uuid

from gcl_sdk.agents.universal.dm import models


class TargetFieldItem(tp.NamedTuple):
    kind: str
    uuid: sys_uuid.UUID
    fields: frozenset[str]

    @classmethod
    def from_ua_resource(cls, resource: models.Resource) -> TargetFieldItem:
        return cls(
            resource.kind, resource.uuid, frozenset(resource.value.keys())
        )


class AbstractTargetFieldsStorage(abc.ABC):
    """Abstract target fields storage.

    Abstract class that represents a storage for target fields.
    UUID of an item is unique across the kind.
    """

    @abc.abstractmethod
    def get(self, kind: str, uuid: sys_uuid.UUID) -> TargetFieldItem:
        """Get the target fields item from the storage."""

    @abc.abstractmethod
    def create(
        self,
        item: TargetFieldItem,
        force: bool = False,
    ) -> TargetFieldItem:
        """Creates the target fields item in the storage."""

    @abc.abstractmethod
    def update(self, item: TargetFieldItem) -> TargetFieldItem:
        """Update the target fields item in the storage."""

    @abc.abstractmethod
    def list(self, kind: str) -> list[TargetFieldItem]:
        """Lists all target fields items of a resource kind."""

    @abc.abstractmethod
    def delete(self, item: TargetFieldItem, force: bool = False) -> None:
        """Delete the target fields item from the storage."""

    @abc.abstractmethod
    def load(self) -> None:
        """Load the storage."""

    @abc.abstractmethod
    def persist(self) -> None:
        """Persist the storage."""



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import uuid as sys_uuid

import bazooka

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.clients.http import base


class UniversalAgentsClient(base.StaticCollectionBaseModelClient):
    __collection_path__ = "/v1/agents/"
    __model__ = models.UniversalAgent

    def get_payload(
        self, uuid: sys_uuid.UUID, last_payload: models.Payload
    ) -> models.Payload:
        payload_data = self.do_action(
            "get_payload",
            uuid,
            hash=last_payload.hash,
            version=last_payload.version,
        )
        cp_payload = models.Payload.restore_from_simple_view(**payload_data)

        # Choose the target payload. If the payloads are equal, the CP returns
        # light payload without capabilities so use the last payload
        # in this case.
        return last_payload if last_payload == cp_payload else cp_payload


class OrchAPI:
    def __init__(
        self,
        base_url: str,
        http_client: bazooka.Client | None = None,
    ) -> None:
        http_client = http_client or bazooka.Client()

        self._http_client = http_client
        self._agents_client = UniversalAgentsClient(base_url, http_client)

    @property
    def agents(self):
        return self._agents_client



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import uuid as sys_uuid

from gcl_sdk.agents.universal.storage import base
from gcl_sdk.agents.universal.storage import exceptions as se
from gcl_sdk.agents.universal.storage import common


class TargetFieldsFileStorage(base.AbstractTargetFieldsStorage):
    """Target fields JSON file storage.

    It stores the target fields in a JSON file.
    The file structure is the following:
    {kind: {uuid: fields}}
    """

    def __init__(self, storage_path: str) -> None:
        self._storage = common.JsonFileStorageSingleton(storage_path)

    def get(self, kind: str, uuid: sys_uuid.UUID) -> base.TargetFieldItem:
        """Get the target fields item from the storage."""
        try:
            fields = self._storage[kind][str(uuid)]
        except KeyError:
            raise se.ItemNotFound(
                item=base.TargetFieldItem(kind, uuid, frozenset())
            )

        return base.TargetFieldItem(kind, uuid, frozenset(fields))

    def create(
        self,
        item: base.TargetFieldItem,
        force: bool = False,
    ) -> base.TargetFieldItem:
        """Creates the target fields item in the storage."""
        try:
            self.get(item.kind, item.uuid)
        except se.ItemNotFound:
            # Desirable behavior, the item should not exist
            pass
        else:
            if not force:
                raise se.ItemAlreadyExists(item=item)

        self._storage.setdefault(item.kind, {})[str(item.uuid)] = list(
            item.fields
        )
        return item

    def update(self, item: base.TargetFieldItem) -> base.TargetFieldItem:
        """Update the target fields item in the storage."""
        return self.create(item, force=True)

    def list(self, kind: str) -> list[base.TargetFieldItem]:
        """Lists all target fields items of a resource kind."""
        return [
            base.TargetFieldItem(kind, sys_uuid.UUID(uuid), frozenset(fields))
            for uuid, fields in self._storage.get(kind, {}).items()
        ]

    def delete(self, item: base.TargetFieldItem, force: bool = False) -> None:
        """Delete the target fields item from the storage."""
        try:
            self.get(item.kind, item.uuid)
        except se.ItemNotFound:
            if not force:
                raise
        else:
            self._storage[item.kind].pop(str(item.uuid), None)

    def load(self) -> None:
        """Load the storage."""
        # Nothing to do. It is loaded on init.
        pass

    def persist(self) -> None:
        """Persist the storage."""
        self._storage.persist()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import logging
import typing as tp

from gcl_sdk.agents.universal.dm import models as ua_models
from gcl_sdk.agents.universal.services import builder


LOG = logging.getLogger(__name__)


class InfraResourcePair(tp.NamedTuple):
    target: ua_models.TargetResourceKindAwareMixin
    actual: ua_models.TargetResourceKindAwareMixin | None


class InfraCollection(tp.NamedTuple):
    infra_objects: tp.Collection[InfraResourcePair]

    def targets(self) -> tuple[ua_models.TargetResourceKindAwareMixin, ...]:
        return tuple(p.target for p in self.infra_objects)

    def actuals(self) -> tuple[ua_models.TargetResourceKindAwareMixin, ...]:
        return tuple(p.actual for p in self.infra_objects)


class CoreInfraBuilder(builder.UniversalBuilderService):

    def create_infra(
        self, instance: ua_models.InstanceWithDerivativesMixin
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Create a list of infrastructure objects.

        The method returns a list of infrastructure objects that are required
        for the instance. For example, nodes, sets, configs, etc.
        """
        raise NotImplementedError

    def actualize_infra(
        self,
        instance: ua_models.InstanceWithDerivativesMixin,
        infra: InfraCollection,
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Actualize the infrastructure objects.

        The method is called when the instance is outdated. For example,
        the instance `Config` has derivative `Render`. Single `Config` may
        have multiple `Render` derivatives. If any of the derivatives is
        outdated, this method is called to reactualize this infrastructure.

        Args:
            instance: The instance to actualize.
            infra: The infrastructure objects.
        """
        return infra.targets()

    def create_instance_derivatives(
        self, instance: ua_models.InstanceWithDerivativesMixin
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Create the instance.

        The result is an iterable object of derivative objects that are
        required for the instance. For example, the main instance is a
        `Config` so the derivative objects for the the config is a list
        of `Render`. The result is a tuple/list/set/... of render objects.
        The derivative objects should inherit from the `TargetResourceMixin`.
        """
        return self.create_infra(instance)

    def update_instance_derivatives(
        self,
        instance: ua_models.InstanceWithDerivativesMixin,
        resource: ua_models.TargetResourceKindAwareMixin,
        derivative_pairs: tp.Collection[
            tuple[
                ua_models.TargetResourceKindAwareMixin,  # The target resource
                ua_models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """The hook to update instance derivatives.

        The hook is called when an initiator of updating is an user or
        software from control plane side.
        The default behavior is to send the same list as on instance creation.
        """
        infra_collection = InfraCollection(
            infra_objects=tuple(
                InfraResourcePair(*p) for p in derivative_pairs
            ),
        )

        return self.actualize_infra(instance, infra_collection)

    def actualize_outdated_instance_derivatives(
        self,
        instance: ua_models.InstanceMixin,
        derivative_pairs: tp.Collection[
            tuple[
                ua_models.TargetResourceKindAwareMixin,  # The target resource
                ua_models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Actualize outdated instance with derivatives.

        It means some changes occurred on the data plane and the instance
        is outdated now. For example, the instance `Config` has derivative
        `Render`. Single `Config` may have multiple `Render` derivatives.
        If any of the derivatives is outdated, this method is called to
        reactualize the derivatives. The method returns the list of `updated`
        derivatives. If nothing needs to be updated, the method returns the
        same list of target derivatives as it received. Otherwise, the method
        should return the list of updated derivatives. It also can add new or
        remove old derivatives.
        Depends on the `fetch_all_derivatives_on_outdate` the behavior of the
        method is different:

        fetch_all_derivatives_on_outdate == True:
        The method receives the list of all derivatives currently available
        for the instance even though the derivatives aren't outdated.

        fetch_all_derivatives_on_outdate == False:
        The method receives the list only changed derivatives from the last
        actualization. For example, a config has two renders. Only one of
        them is outdated. The method receives the list of only one outdated
        render.

        Args:
            instance: The instance to actualize.
            derivative_pairs: Changed or all derivatives of the instance.
        """
        infra_collection = InfraCollection(
            infra_objects=tuple(
                InfraResourcePair(*p) for p in derivative_pairs
            ),
        )

        return self.actualize_infra(instance, infra_collection)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstarctMigrationStep):

    def __init__(self):
        self._depends = ["0002-init-auditlog-table-c6f740.py"]

    @property
    def migration_id(self):
        return "6e9ca843-3268-4449-a507-76258f43c53f"

    @property
    def is_manual(self):
        return False

    def upgrade(self, session):
        sql_expressions = [
            """
            ALTER TABLE ua_target_resources
                ADD COLUMN master_hash varchar(256) NOT NULL DEFAULT '';
            """,
            """
            ALTER TABLE ua_target_resources
                ADD COLUMN master_full_hash varchar(256) NOT NULL DEFAULT '';
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_target_resources_hash_idx
                ON ua_target_resources (hash);
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_target_resources_full_hash_idx
                ON ua_target_resources (full_hash);
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_target_resources_master_hash_idx
                ON ua_target_resources (master_hash);
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_target_resources_master_full_hash_idx
                ON ua_target_resources (master_full_hash);
            """,
            """
            CREATE OR REPLACE VIEW ua_outdated_master_hash_resources_view AS
                SELECT
                    ua_target_resources.uuid as uuid,
                    ua_target_resources.kind as kind,
                    ua_target_resources.res_uuid as target_resource,
                    masters.res_uuid as master
                FROM ua_target_resources JOIN ua_target_resources as masters ON 
                    ua_target_resources.master = masters.uuid
                WHERE ua_target_resources.master_hash != masters.hash AND
                    ua_target_resources.kind != masters.kind;
            """,
            """
            CREATE OR REPLACE VIEW ua_outdated_master_full_hash_resources_view AS
                SELECT
                    ua_target_resources.uuid as uuid,
                    ua_target_resources.kind as kind,
                    ua_target_resources.res_uuid as target_resource,
                    masters.res_uuid as master
                FROM ua_target_resources JOIN ua_target_resources as masters ON 
                    ua_target_resources.master = masters.uuid
                WHERE ua_target_resources.master_full_hash != masters.full_hash AND
                    ua_target_resources.kind != masters.kind;
            """,
        ]

        for expr in sql_expressions:
            session.execute(expr, None)

    def downgrade(self, session):
        sql_types = [
            "ALTER TABLE ua_target_resources DROP COLUMN IF EXISTS master_hash;",
            "ALTER TABLE ua_target_resources DROP COLUMN IF EXISTS master_full_hash;",
            "DROP INDEX IF EXISTS ua_target_resources_hash_idx;",
            "DROP INDEX IF EXISTS ua_target_resources_full_hash_idx;",
            "DROP INDEX IF EXISTS ua_target_resources_master_hash_idx;",
            "DROP INDEX IF EXISTS ua_target_resources_master_full_hash_idx;",
        ]

        views = [
            "ua_outdated_master_hash_resources_view",
            "ua_outdated_master_full_hash_resources_view",
        ]

        for view_name in views:
            self._delete_view_if_exists(session, view_name)

        for expr in sql_types:
            session.execute(expr, None)


migration_step = MigrationStep()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.api import routes

from gcl_sdk.agents.universal.status_api import controllers


class UniversalAgentsRoute(routes.Route):
    """Handler for /v1/agents/ endpoint"""

    __controller__ = controllers.UniversalAgentsController


class ResourcesRoute(routes.Route):
    """Handler for /v1/kind/<name>/resources/ endpoint"""

    __controller__ = controllers.ResourcesController


class KindRoute(routes.Route):
    """Handler for /v1/kind/ endpoint"""

    __allow_methods__ = [routes.FILTER, routes.GET]
    __controller__ = controllers.KindController

    resources = routes.route(ResourcesRoute, resource_route=True)



import uuid as sys_uuid

from gcl_iam import controllers as iam_controllers
from restalchemy.api import controllers, resources

from gcl_sdk.audit.dm import models


class AuditController(
    iam_controllers.PolicyBasedControllerMixin,
    controllers.BaseResourceControllerPaginated,
):
    __policy_service_name__ = "audit_log"
    __policy_name__ = "audit_record"
    __resource__ = resources.ResourceByRAModel(
        model_class=models.AuditRecord,
        process_filters=True,
        convert_underscore=False,
    )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import abc
import typing as tp

from gcl_sdk.agents.universal.dm import models


class AbstractBackendClient(abc.ABC):
    """Abstract backend client.

    An interface for clients to communicate with a backend side.
    """

    @abc.abstractmethod
    def get(
        self, resource: models.Resource
    ) -> dict[str, tp.Any] | models.Resource | models.ResourceMixin:
        """Get the resource.

        The method can return resource in different formats:
         - dictionary - raw data of the resource value
         - Resource - ready to use resource model
         - ResourceMixin - a custom model that extends ResourceMixin and
            can be converted to Resource model.
        """

    @abc.abstractmethod
    def create(
        self, resource: models.Resource
    ) -> dict[str, tp.Any] | models.Resource | models.ResourceMixin:
        """Creates the resource. Returns the created resource.

        The method can return resource in different formats:
         - dictionary - raw data of the resource value
         - Resource - ready to use resource model
         - ResourceMixin - a custom model that extends ResourceMixin and
            can be converted to Resource model.
        """

    @abc.abstractmethod
    def update(
        self, resource: models.Resource
    ) -> dict[str, tp.Any] | models.Resource | models.ResourceMixin:
        """Update the resource. Returns the updated resource.

        The method can return resource in different formats:
         - dictionary - raw data of the resource value
         - Resource - ready to use resource model
         - ResourceMixin - a custom model that extends ResourceMixin and
            can be converted to Resource model.
        """

    @abc.abstractmethod
    def list(
        self, kind: str, **kwargs
    ) -> tp.Collection[
        dict[str, tp.Any] | models.Resource | models.ResourceMixin
    ]:
        """Lists all resources by kind.

        The method returns collection of resources in different formats:
         - dictionary - raw data of the resource value
         - Resource - ready to use resource model
         - ResourceMixin - a custom model that extends ResourceMixin and
            can be converted to Resource model.
        """

    @abc.abstractmethod
    def delete(self, resource: models.Resource) -> None:
        """Delete the resource."""



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstarctMigrationStep):

    def __init__(self):
        self._depends = ["0000-init-events-table-2cfd220e.py"]

    @property
    def migration_id(self):
        return "391c09b0-9093-4a82-a310-34e2b65bf6c7"

    @property
    def is_manual(self):
        return False

    def upgrade(self, session):
        sql_expressions = [
            # TABLES
            """
            DROP TYPE IF EXISTS enum_agent_status;
            CREATE TYPE "enum_agent_status" AS ENUM (
                'NEW',
                'ACTIVE',
                'ERROR',
                'DISABLED'
            );
            """,
            """
            CREATE TABLE IF NOT EXISTS ua_agents (
                "uuid" UUID NOT NULL PRIMARY KEY,
                "name" varchar(255) NOT NULL,
                "description" varchar(255) NOT NULL,
                "capabilities" JSONB NOT NULL,
                "facts" JSONB NOT NULL,
                "node" UUID NOT NULL,
                "status" enum_agent_status NOT NULL DEFAULT 'NEW',
                "created_at" timestamp NOT NULL DEFAULT current_timestamp,
                "updated_at" timestamp NOT NULL DEFAULT current_timestamp
            );
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_agents_node_idx
                ON ua_agents (node);
            """,
            """
            CREATE TABLE IF NOT EXISTS ua_actual_resources (
                "uuid" UUID NOT NULL,
                "kind" varchar(64) NOT NULL,
                "res_uuid" UUID NOT NULL PRIMARY KEY,
                "value" JSONB NOT NULL,
                "status" varchar(32) NOT NULL,
                "node" UUID DEFAULT NULL,
                "hash" varchar(256) NOT NULL,
                "full_hash" varchar(256) NOT NULL,
                "created_at" timestamp NOT NULL DEFAULT current_timestamp,
                "updated_at" timestamp NOT NULL DEFAULT current_timestamp,
                UNIQUE (uuid, kind)
            );
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_actual_resources_node_id_idx
                ON ua_actual_resources (node);
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_actual_resources_kind_id_idx
                ON ua_actual_resources (kind);
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_actual_resources_full_hash_id_idx
                ON ua_actual_resources (full_hash);
            """,
            """
            CREATE TABLE IF NOT EXISTS ua_target_resources (
                "uuid" UUID NOT NULL,
                "kind" varchar(64) NOT NULL,
                "res_uuid" UUID NOT NULL PRIMARY KEY,
                "value" JSONB NOT NULL,
                "status" varchar(32) NOT NULL DEFAULT 'ACTIVE',
                "hash" varchar(256) NOT NULL,
                "full_hash" varchar(256) NOT NULL,
                "node" UUID DEFAULT NULL,
                "agent" UUID references ua_agents(uuid) ON DELETE CASCADE,
                "master" UUID DEFAULT NULL,
                "tracked_at" timestamp NOT NULL DEFAULT current_timestamp,
                "created_at" timestamp NOT NULL DEFAULT current_timestamp,
                "updated_at" timestamp NOT NULL DEFAULT current_timestamp,
                UNIQUE (uuid, kind)
            );
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_target_resources_agent_id_idx
                ON ua_target_resources (agent);
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_target_resources_master_id_idx
                ON ua_target_resources (master);
            """,
            """
            CREATE INDEX IF NOT EXISTS ua_target_resources_kind_id_idx
                ON ua_target_resources (kind);
            """,
            """
            CREATE OR REPLACE VIEW ua_outdated_resources_view AS
                SELECT
                    ua_target_resources.uuid as uuid,
                    ua_target_resources.kind as kind,
                    ua_target_resources.res_uuid as target_resource,
                    ua_actual_resources.res_uuid as actual_resource
                FROM ua_target_resources INNER JOIN ua_actual_resources ON 
                    ua_target_resources.uuid = ua_actual_resources.uuid
                WHERE ua_target_resources.full_hash != ua_actual_resources.full_hash;
            """,
        ]

        for expr in sql_expressions:
            session.execute(expr, None)

    def downgrade(self, session):
        sql_types = [
            """
            DROP TYPE IF EXISTS enum_agent_status;
            """,
        ]

        tables = [
            "ua_actual_resources",
            "ua_target_resources",
            "ua_agents",
        ]

        views = [
            "ua_outdated_resources_view",
        ]

        for view_name in views:
            self._delete_view_if_exists(session, view_name)

        for table_name in tables:
            self._delete_table_if_exists(session, table_name)

        for expr in sql_types:
            session.execute(expr, None)


migration_step = MigrationStep()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import os

from gcl_sdk.agents.universal.drivers import base
from gcl_sdk.agents.universal.drivers import exceptions as driver_exc
from gcl_sdk.agents.universal.dm import models


class DummyFilesDriver(base.AbstractCapabilityDriver):
    """An example driver for testing.

    This driver is intended for testing and development only.
    """

    def __init__(self, work_dir: str):
        super().__init__()
        self._work_dir = work_dir

    def get(self, resource: models.Resource) -> models.Resource:
        """Find and return the file resource."""
        for f in os.listdir(self.work_dir):
            uuid, name = f.split("_", 1)
            if uuid == str(resource.uuid):
                value = {"uuid": uuid, "name": name}
                new_resource = models.Resource.from_value(value, resource.kind)
                return new_resource

        raise driver_exc.ResourceNotFound(resource=resource)

    def list(self, capability: str) -> list[models.Resource]:
        """Lists all files"""
        resources = []

        for f in os.listdir(self.work_dir):
            uuid, name = f.split("_", 1)
            value = {"uuid": uuid, "name": name}
            resource = models.Resource.from_value(value, capability)
            resources.append(resource)

        return resources

    def create(self, resource: models.Resource) -> models.Resource:
        """Creates a new file in the work directory."""
        name = f"{resource.value['uuid']}_{resource.value['name']}"
        path = os.path.join(self.work_dir, name)
        with open(path, "w") as f:
            f.write("")
        return resource

    def update(self, resource: models.Resource) -> models.Resource:
        """Update the file in the work directory."""
        self.delete(resource)
        return self.create(resource)

    def delete(self, resource: models.Resource) -> None:
        """Delete the file in the work directory."""
        try:
            res = self.get(resource)
        except driver_exc.ResourceNotFound:
            # Nothing to do, the resource does not exist
            return

        name = f"{res.value['uuid']}-{res.value['name']}"
        path = os.path.join(self.work_dir, name)
        os.remove(path)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import tempfile
import uuid as sys_uuid
from unittest import mock

from gcl_sdk.agents.universal import utils
from gcl_sdk.agents.universal.clients.orch import exceptions as orch_exc
from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal.drivers import base
from gcl_sdk.agents.universal.services import agent as agent_svc
from gcl_sdk.tests.functional import conftest

MACHINE_UUID = sys_uuid.UUID("44b5857b-c15d-47f2-bed1-00fecd137208")


class FooCapDriver(base.AbstractCapabilityDriver):
    create_called = False
    delete_called = False
    update_called = False
    list_called = False
    get_called = False

    def get_capabilities(self) -> list[str]:
        return ["foo"]

    def get(self, resource: models.Resource) -> models.Resource:
        self.__class__.get_called = True
        return resource

    def create(self, resource: models.Resource) -> models.Resource:
        self.__class__.create_called = True
        return resource

    def update(self, resource: models.Resource) -> models.Resource:
        self.__class__.update_called = True
        return resource

    def list(self, capability: str) -> list[models.Resource]:
        self.__class__.list_called = True
        return []

    def delete(self, resource: models.Resource) -> None:
        self.__class__.delete_called = True
        pass


class FooFactDriver(base.AbstractFactDriver):
    get_called = False
    list_called = False

    def get_facts(self) -> list[str]:
        return ["foo"]

    def get(self, resource: models.Resource) -> models.Resource:
        self.__class__.get_called = True
        return resource

    def list(self, fact: str) -> list[models.Resource]:
        self.__class__.list_called = True
        return []


@mock.patch.object(
    utils,
    "system_uuid",
    lambda: MACHINE_UUID,
)
class TestUniversalAgent:
    PAYLOAD_PATH = os.path.join(tempfile.gettempdir(), "___payload___.json")

    def setup_method(self) -> None:
        # Run service
        self.orch_client = mock.MagicMock()
        self.agent_uuid = MACHINE_UUID

    def teardown_method(self) -> None:
        if os.path.exists(self.PAYLOAD_PATH):
            os.remove(self.PAYLOAD_PATH)

    def test_agent_registration(self):
        # Simulate agent not found on orchestrator
        self.orch_client.agents_get_payload = mock.MagicMock(
            side_effect=orch_exc.AgentNotFound(uuid=self.agent_uuid)
        )

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[],
            facts_drivers=[],
        )

        agent._iteration()

        self.orch_client.agents_create.assert_called_once()

    def test_agent_new_capability(self):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(uuid=uuid)
        payload = models.Payload.empty()
        payload.add_caps_resource(resource.to_ua_resource(kind="foo"))
        payload.calculate_hash()

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class CapDriver(FooCapDriver):

            def create(self, resource):
                assert resource.uuid == uuid
                return super().create(resource)

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[CapDriver()],
            facts_drivers=[],
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        assert CapDriver.create_called
        assert not CapDriver.delete_called
        assert not CapDriver.update_called
        self.orch_client.agents_get_payload.assert_called_once()
        # A fact for created capability should be created
        # in Status via orch client
        self.orch_client.resources_create.assert_called_once()
        created_res = self.orch_client.resources_create.call_args.args[0]
        assert isinstance(created_res, models.Resource)
        assert created_res.uuid == uuid
        assert created_res.kind == "foo"
        self.orch_client.resources_update.assert_not_called()
        self.orch_client.resources_delete.assert_not_called()

    def test_agent_update_capability(self):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(uuid=uuid)
        payload = models.Payload.empty()
        payload.add_caps_resource(resource.to_ua_resource(kind="foo"))
        payload.add_facts_resource(resource.to_ua_resource(kind="foo"))
        payload.calculate_hash()

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class CapDriver(FooCapDriver):
            def list(self, capability):
                res = resource.dump_to_simple_view()
                res["name"] = "foo-name-updated"
                res = conftest.FooResource.restore_from_simple_view(**res)
                return [res.to_ua_resource(kind="foo")]

            def update(self, resource):
                assert resource.uuid == uuid
                return super().update(resource)

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[CapDriver()],
            facts_drivers=[],
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        assert CapDriver.update_called
        assert not CapDriver.delete_called
        assert not CapDriver.create_called
        self.orch_client.agents_get_payload.assert_called_once()
        self.orch_client.resources_create.assert_not_called()
        self.orch_client.resources_update.assert_not_called()
        self.orch_client.resources_delete.assert_not_called()

    def test_agent_delete_capability(self):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(uuid=uuid)
        payload = models.Payload.empty()
        payload.capabilities = {"foo": {"resources": []}}
        payload.add_facts_resource(resource.to_ua_resource(kind="foo"))
        payload.calculate_hash()

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class CapDriver(FooCapDriver):
            def list(self, capability):
                return [resource.to_ua_resource(kind="foo")]

            def delete(self, resource):
                assert resource.uuid == uuid
                return super().delete(resource)

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[CapDriver()],
            facts_drivers=[],
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        assert CapDriver.delete_called
        assert not CapDriver.update_called
        assert not CapDriver.create_called
        self.orch_client.agents_get_payload.assert_called_once()
        self.orch_client.resources_create.assert_not_called()
        self.orch_client.resources_update.assert_not_called()
        self.orch_client.resources_delete.assert_called_once()
        deleted_res = self.orch_client.resources_delete.call_args.args[0]
        assert isinstance(deleted_res, models.Resource)
        assert deleted_res.uuid == uuid

    def test_agent_new_capabilities(self):
        uuid_a = sys_uuid.uuid4()
        uuid_b = sys_uuid.uuid4()
        resource_a = conftest.FooResource(uuid=uuid_a)
        resource_b = conftest.FooResource(uuid=uuid_b)
        payload = models.Payload.empty()
        payload.add_caps_resource(resource_a.to_ua_resource(kind="foo"))
        payload.add_caps_resource(resource_b.to_ua_resource(kind="foo"))
        payload.calculate_hash()

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class CapDriver(FooCapDriver):

            def create(self, resource):
                assert resource.uuid in {uuid_a, uuid_b}
                return super().create(resource)

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[CapDriver()],
            facts_drivers=[],
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        self.orch_client.agents_get_payload.assert_called_once()
        assert self.orch_client.resources_create.call_count == 2
        self.orch_client.resources_update.assert_not_called()
        self.orch_client.resources_delete.assert_not_called()

    def test_agent_mix_capability_actions(self):
        uuid_a = sys_uuid.uuid4()
        uuid_b = sys_uuid.uuid4()
        uuid_c = sys_uuid.uuid4()
        resource_a = conftest.FooResource(uuid=uuid_a)
        resource_b = conftest.FooResource(uuid=uuid_b)
        resource_c = conftest.FooResource(uuid=uuid_c)
        payload = models.Payload.empty()
        payload.add_caps_resource(resource_a.to_ua_resource(kind="foo"))
        payload.add_caps_resource(resource_b.to_ua_resource(kind="foo"))
        payload.add_facts_resource(resource_b.to_ua_resource(kind="foo"))
        payload.add_facts_resource(resource_c.to_ua_resource(kind="foo"))
        payload.calculate_hash()

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class CapDriver(FooCapDriver):

            def list(self, capability):
                res = resource_b.dump_to_simple_view()
                res["name"] = "foo-name-updated"
                res = conftest.FooResource.restore_from_simple_view(**res)
                return [
                    res.to_ua_resource(kind="foo"),
                    resource_c.to_ua_resource(kind="foo"),
                ]

            def create(self, resource):
                assert resource.uuid == uuid_a
                return super().create(resource)

            def delete(self, resource):
                assert resource.uuid == uuid_c
                return super().delete(resource)

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[CapDriver()],
            facts_drivers=[],
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        self.orch_client.resources_create.assert_called_once()
        created_res = self.orch_client.resources_create.call_args.args[0]
        assert isinstance(created_res, models.Resource)
        assert created_res.uuid == uuid_a
        self.orch_client.resources_update.assert_not_called()
        self.orch_client.resources_delete.assert_called_once()
        deleted_res = self.orch_client.resources_delete.call_args.args[0]
        assert isinstance(deleted_res, models.Resource)
        assert deleted_res.uuid == uuid_c

    def test_agent_new_fact(self):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(uuid=uuid)
        payload = models.Payload.empty()
        payload.facts = {"foo": {"resources": []}}
        payload.calculate_hash()

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class FactDriver(FooFactDriver):

            def list(self, fact):
                super().list(fact)
                return [resource.to_ua_resource(kind="foo")]

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[],
            facts_drivers=[FactDriver()],
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        self.orch_client.agents_get_payload.assert_called_once()
        self.orch_client.resources_create.assert_called_once()
        created_res = self.orch_client.resources_create.call_args.args[0]
        assert isinstance(created_res, models.Resource)
        assert created_res.uuid == uuid
        assert created_res.kind == "foo"
        self.orch_client.resources_update.assert_not_called()
        self.orch_client.resources_delete.assert_not_called()

    def test_agent_update_fact(self):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(uuid=uuid)
        payload = models.Payload.empty()
        payload.add_facts_resource(resource.to_ua_resource(kind="foo"))
        payload.calculate_hash()

        res = resource.dump_to_simple_view()
        res["name"] = "foo-name-updated"
        res = conftest.FooResource.restore_from_simple_view(**res)

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class FactDriver(FooFactDriver):

            def list(self, fact):
                super().list(fact)
                return [res.to_ua_resource(kind="foo")]

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[],
            facts_drivers=[FactDriver()],
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        self.orch_client.agents_get_payload.assert_called_once()
        self.orch_client.resources_create.assert_not_called()
        self.orch_client.resources_delete.assert_not_called()

        data = res.to_ua_resource(kind="foo").dump_to_simple_view()
        data.pop("created_at", None)
        data.pop("updated_at", None)
        data.pop("res_uuid", None)
        data.pop("node", None)
        uuid = data.pop("uuid")
        kind = data.pop("kind")
        self.orch_client.resources_update.assert_called_once_with(
            kind, uuid, **data
        )

    def test_agent_delete_fact(self):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(uuid=uuid)
        payload = models.Payload.empty()
        payload.add_facts_resource(resource.to_ua_resource(kind="foo"))
        payload.calculate_hash()

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class FactDriver(FooFactDriver):
            pass

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[],
            facts_drivers=[FactDriver()],
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        self.orch_client.agents_get_payload.assert_called_once()
        self.orch_client.resources_create.assert_not_called()
        self.orch_client.resources_update.assert_not_called()
        self.orch_client.resources_delete.assert_called_once()
        deleted_res = self.orch_client.resources_delete.call_args.args[0]
        assert isinstance(deleted_res, models.Resource)
        assert deleted_res.uuid == uuid

    def test_agent_new_facts(self):
        uuid_a = sys_uuid.uuid4()
        uuid_b = sys_uuid.uuid4()
        resource_a = conftest.FooResource(uuid=uuid_a)
        resource_b = conftest.FooResource(uuid=uuid_b)
        payload = models.Payload.empty()
        payload.facts = {"foo": {"resources": []}}
        payload.calculate_hash()

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class FactDriver(FooFactDriver):

            def list(self, fact):
                super().list(fact)
                return [
                    resource_a.to_ua_resource(kind="foo"),
                    resource_b.to_ua_resource(kind="foo"),
                ]

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            orch_client=self.orch_client,
            caps_drivers=[],
            facts_drivers=[FactDriver()],
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        self.orch_client.agents_get_payload.assert_called_once()
        assert self.orch_client.resources_create.call_count == 2
        self.orch_client.resources_update.assert_not_called()
        self.orch_client.resources_delete.assert_not_called()

    def test_agent_mix_fact_actions(self):
        uuid_a = sys_uuid.uuid4()
        uuid_b = sys_uuid.uuid4()
        uuid_c = sys_uuid.uuid4()
        resource_a = conftest.FooResource(uuid=uuid_a)
        resource_b = conftest.FooResource(uuid=uuid_b)
        resource_c = conftest.FooResource(uuid=uuid_c)
        payload = models.Payload.empty()
        payload.add_facts_resource(resource_b.to_ua_resource(kind="foo"))
        payload.add_facts_resource(resource_c.to_ua_resource(kind="foo"))
        payload.calculate_hash()

        res = resource_b.dump_to_simple_view()
        res["name"] = "foo-name-updated"
        res = conftest.FooResource.restore_from_simple_view(**res)

        self.orch_client.agents_get_payload = mock.MagicMock(
            return_value=payload
        )

        class FactDriver(FooFactDriver):

            def list(self, fact):
                super().list(fact)
                return [
                    resource_a.to_ua_resource(kind="foo"),
                    res.to_ua_resource(kind="foo"),
                ]

        agent = agent_svc.UniversalAgentService(
            agent_uuid=self.agent_uuid,
            caps_drivers=[],
            facts_drivers=[FactDriver()],
            orch_client=self.orch_client,
            payload_path=self.PAYLOAD_PATH,
        )

        agent._iteration()

        self.orch_client.agents_get_payload.assert_called_once()
        self.orch_client.resources_create.assert_called_once()
        created_res = self.orch_client.resources_create.call_args.args[0]
        assert isinstance(created_res, models.Resource)
        assert created_res.uuid == uuid_a
        self.orch_client.resources_delete.assert_called_once()
        deleted_res = self.orch_client.resources_delete.call_args.args[0]
        assert isinstance(deleted_res, models.Resource)
        assert deleted_res.uuid == uuid_c

        data = res.to_ua_resource(kind="foo").dump_to_simple_view()
        data.pop("created_at", None)
        data.pop("updated_at", None)
        data.pop("res_uuid", None)
        data.pop("node", None)
        uuid = data.pop("uuid")
        kind = data.pop("kind")
        self.orch_client.resources_update.assert_called_once_with(
            kind, uuid, **data
        )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import abc

from gcl_sdk.agents.universal.dm import models


class AbstractCapabilityDriver(abc.ABC):
    """Abstract driver for capabilities.

    The main purpose of this driver is to apply target
    capabilities to the DP.
    """

    @abc.abstractmethod
    def get_capabilities(self) -> list[str]:
        """Returns a list of capabilities supported by the driver."""

    @abc.abstractmethod
    def get(self, resource: models.Resource) -> models.Resource:
        """Find and return a resource by uuid and kind.

        It returns the resource from the data plane.
        """

    @abc.abstractmethod
    def create(self, resource: models.Resource) -> models.Resource:
        """Creates a resource."""

    @abc.abstractmethod
    def update(self, resource: models.Resource) -> models.Resource:
        """Update the resource.

        The simplest implementation. The driver should detect which
        fields were changed itself.
        """

    @abc.abstractmethod
    def list(self, capability: str) -> list[models.Resource]:
        """Lists all resources by capability."""

    @abc.abstractmethod
    def delete(self, resource: models.Resource) -> None:
        """Delete the resource."""

    def start(self) -> None:
        """Perform some initialization before starting any operations.

        This method is called once before any other method like list,
        create, update, delete are called. It can be used to do some
        preparations like establishing connections, opening files, etc.

        The driver iteration:
            start -> list -> [create | update | delete]* -> finalize
        """
        pass

    def finalize(self) -> None:
        """Perform some finalization after finishing all operations.

        This method is called once after all other methods like list,
        create, update, delete are called. It can be used to do some
        finalization or cleanups like closing connections, files, etc.

        The driver iteration:
            start -> list -> [create | update | delete]* -> finalize
        """
        pass


class AbstractFactDriver(abc.ABC):
    """Abstract driver for facts.

    The main purpose of this driver is to gather facts from the
    data plane. The simplest example is gathering network interfaces
    from nodes.
    """

    @abc.abstractmethod
    def get_facts(self) -> list[str]:
        """Returns a list of facts supported by the driver."""

    @abc.abstractmethod
    def get(self, resource: models.Resource) -> models.Resource:
        """Find and return a resource by uuid and kind.

        It returns the resource from the data plane.
        """

    @abc.abstractmethod
    def list(self, fact: str) -> list[models.Resource]:
        """Lists all resources by facts."""



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstarctMigrationStep):

    def __init__(self):
        self._depends = ["0001-universal-agent-391c09.py"]

    @property
    def migration_id(self):
        return "c6f74019-47cb-412a-a11a-d42921f694a4"

    @property
    def is_manual(self):
        return False

    def upgrade(self, session):
        expressions = [
            """
            CREATE TABLE IF NOT EXISTS "gcl_sdk_audit_logs" (
                "uuid" UUID PRIMARY KEY,
                "object_uuid" UUID NOT NULL,
                "object_type" varchar(64) NOT NULL,
                "user_uuid" UUID DEFAULT NULL,
                "action" varchar(64) NOT NULL,
                "created_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "updated_at" TIMESTAMP(6) NOT NULL DEFAULT NOW()
            );
            """,
            """
            CREATE INDEX IF NOT EXISTS gcl_sdk_audit_logs_object_type_action_idx
                ON gcl_sdk_audit_logs (object_type, action);
            """,
        ]

        for expression in expressions:
            session.execute(expression)

    def downgrade(self, session):
        tables = ["gcl_sdk_audit_logs"]

        for table in tables:
            self._delete_table_if_exists(session, table)


migration_step = MigrationStep()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from restalchemy.api import actions
from restalchemy.api import resources
from restalchemy.api import controllers

from gcl_sdk.agents.universal.dm import models


class UniversalAgentsController(controllers.BaseResourceController):
    """Controller for /v1/agents/ endpoint"""

    __resource__ = resources.ResourceByRAModel(
        model_class=models.UniversalAgent,
        process_filters=True,
        convert_underscore=False,
    )

    @actions.get
    def get_payload(
        self,
        resource: models.UniversalAgent,
        hash: str = "",
        version: str = "0",
    ):
        payload = resource.get_payload(hash=hash, version=int(version))
        return payload.dump_to_simple_view()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from gcl_sdk.events import clients
from gcl_sdk.events import constants as event_c


CONF = cfg.CONF


def register_event_opts(conf=None):
    conf = conf or CONF

    events_opts = [
        cfg.BoolOpt(
            "enabled",
            default=True,
            help="Enable event processing",
        ),
        cfg.StrOpt(
            "client_type",
            default="async",
            choices=list(clients.EVENT_CLIENT_MAPPER.keys()),
            help="Event client type",
        ),
        cfg.StrOpt(
            "event_type_mapping_filepath",
            default="event_type_mapping.yaml",
            help="Event type mapping file path",
        ),
        cfg.URIOpt(
            "genesis_notification_endpoint",
            default="http://127.0.0.1:8080/",
            help="Genesis notification endpoint URL",
        ),
        cfg.StrOpt(
            "genesis_api_version",
            choices=["v1"],
            default="v1",
            help="API version of the Genesis service",
        ),
        cfg.StrOpt(
            "genesis_api_token",
            default="<inser token here>",
            help=(
                "API token for authentication with the Genesis"
                " Notificatio service"
            ),
        ),
        cfg.StrOpt(
            "project_id",
            default="00000000-0000-0000-0000-000000000000",
            help="Project ID for the events",
        ),
    ]

    conf.register_cli_opts(events_opts, event_c.DOMAIN)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import logging
import typing as tp
import uuid as sys_uuid

from restalchemy.dm import filters as dm_filters
from restalchemy.storage import exceptions as ra_exc
from restalchemy.storage import base as ra_storage
from gcl_sdk.agents.universal.clients.backend import base
from gcl_sdk.agents.universal.clients.backend import exceptions as client_exc
from gcl_sdk.agents.universal.dm import models


LOG = logging.getLogger(__name__)


class ModelSpec(tp.NamedTuple):
    model: tp.Type[ra_storage.AbstractStorableMixin]
    kind: str

    # TODO(akremenetsky): Actually we can do filtering more flexibly
    # by using any filters from restalchemy and not using the project_id.
    project_id: sys_uuid.UUID

    @classmethod
    def from_collection(
        cls,
        collection: tp.Collection[tuple[tp.Type, str]],
        project_id: sys_uuid.UUID,
    ) -> tuple[ModelSpec, ...]:
        return tuple(
            cls(
                model=model,
                kind=kind,
                project_id=project_id,
            )
            for model, kind in collection
        )


class DatabaseBackendClient(base.AbstractBackendClient):
    """Database backend client."""

    def __init__(
        self,
        model_specs: tp.Collection[ModelSpec],
        session: tp.Any | None = None,
    ):
        super().__init__()
        self._session = session
        self._model_spec_map = {m.kind: m for m in model_specs}

    def set_session(self, session: tp.Any) -> None:
        """Set the session to be used by the client."""
        self._session = session

    def clear_session(self) -> None:
        """Clear the session."""
        self._session = None

    def get(self, resource: models.Resource) -> models.ResourceMixin:
        """Find and return a resource by uuid and kind."""
        model_spec = self._model_spec_map[resource.kind]

        try:
            return model_spec.model.objects.get_one(
                session=self._session,
                filters={
                    "uuid": dm_filters.EQ(str(resource.uuid)),
                    "project_id": dm_filters.EQ(str(model_spec.project_id)),
                },
            )
        except ra_exc.RecordNotFound:
            LOG.exception(
                "Unable to find %s %s", str(model_spec), resource.uuid
            )
            raise client_exc.ResourceNotFound(resource=resource)

    def list(self, capability: str) -> list[models.ResourceMixin]:
        """Lists all resources by capability."""
        model_spec = self._model_spec_map[capability]

        # Get all objects for the project from the database
        return model_spec.model.objects.get_all(
            session=self._session,
            filters={
                "project_id": dm_filters.EQ(str(model_spec.project_id)),
            },
        )

    def create(self, resource: models.Resource) -> models.ResourceMixin:
        """Creates a resource."""
        model_spec = self._model_spec_map[resource.kind]

        # Check if the resource already exists
        # We need to do this check since PG does not correctly handle
        # the case when the resource already exists and fails the transaction
        obj = model_spec.model.objects.get_one_or_none(
            session=self._session,
            filters={
                "uuid": dm_filters.EQ(str(resource.uuid)),
                "project_id": dm_filters.EQ(str(model_spec.project_id)),
            },
        )
        if obj is not None:
            LOG.warning("The resource already exists: %s", resource.uuid)
            raise client_exc.ResourceAlreadyExists(resource=resource)

        # Save to db
        obj = model_spec.model.from_ua_resource(resource)
        obj.insert(session=self._session)

        return obj

    def update(self, resource: models.Resource) -> models.ResourceMixin:
        """Update the resource."""
        model_spec = self._model_spec_map[resource.kind]

        # Check if the resource already exists
        obj = model_spec.model.objects.get_one_or_none(
            session=self._session,
            filters={
                "uuid": dm_filters.EQ(str(resource.uuid)),
                "project_id": dm_filters.EQ(str(model_spec.project_id)),
            },
        )
        if obj is None:
            LOG.warning("The resource does not exist: %s", resource.uuid)
            raise client_exc.ResourceNotFound(resource=resource)

        updated_obj = model_spec.model.from_ua_resource(resource)

        # Update the object
        for field_name in resource.value.keys():
            prop = obj.properties.get(field_name)
            if not prop or prop.is_read_only():
                continue
            setattr(obj, field_name, getattr(updated_obj, field_name))
        obj.update(session=self._session)

        return obj

    def delete(self, resource: models.Resource) -> None:
        """Delete the resource."""
        model_spec = self._model_spec_map[resource.kind]

        try:
            obj = model_spec.model.objects.get_one(
                session=self._session,
                filters={
                    "uuid": dm_filters.EQ(str(resource.uuid)),
                    "project_id": dm_filters.EQ(str(model_spec.project_id)),
                },
            )
            obj.delete(session=self._session)
            LOG.debug("Deleted resource: %s", resource.uuid)
        except ra_exc.RecordNotFound:
            LOG.warning("The resource is already deleted: %s", resource.uuid)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import enum


DEF_ROOT_DISK_SIZE = 10


class InstanceStatus(str, enum.Enum):
    NEW = "NEW"
    IN_PROGRESS = "IN_PROGRESS"
    ACTIVE = "ACTIVE"
    ERROR = "ERROR"


class NodeStatus(str, enum.Enum):
    NEW = "NEW"
    SCHEDULED = "SCHEDULED"
    IN_PROGRESS = "IN_PROGRESS"
    STARTED = "STARTED"
    ACTIVE = "ACTIVE"
    ERROR = "ERROR"


class NodeType(str, enum.Enum):
    VM = "VM"
    HW = "HW"


class NodeSetType(str, enum.Enum):
    SET = "SET"



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import logging
import uuid as sys_uuid

from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.dm import models as ra_models

from gcl_sdk.agents.universal.drivers import base
from gcl_sdk.agents.universal.drivers import exceptions as driver_exc
from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal.storage import common as storage_common


LOG = logging.getLogger(__name__)


class MetaDataPlaneModel(
    ra_models.ModelWithRequiredUUID, models.ResourceMixin
):
    """A base model for using in MetaFileStorageAgentDriver.

    Child models should implement methods to work with data plane.
    """

    # Store the resource target fields
    target_fields = properties.property(
        types.TypedList(types.String()), default=list
    )

    @classmethod
    def from_ua_resource(
        cls, resource: models.Resource
    ) -> "MetaDataPlaneModel":
        target_fields = list(resource.value.keys())
        return cls.restore_from_simple_view(
            target_fields=target_fields, **resource.value
        )

    def get_resource_target_fields(self) -> list[str]:
        """Return the list of target fields.

        Refer to the Resource model for more details about target fields.
        """
        return self.target_fields

    def get_resource_ignore_fields(self) -> list[str]:
        """Return fields that should not belong to the resource."""
        return ["target_fields"]

    def get_meta_fields(self) -> set[str] | None:
        """Return a list of meta fields or None.

        Similar to `get_meta_model_fields` but adds addtional service fields.
        """
        fields = self.get_meta_model_fields()
        if fields is None:
            return None
        fields.add("target_fields")
        return fields

    def get_meta_model_fields(self) -> set[str] | None:
        """Return a list of meta fields or None.

        Meta fields are the fields that cannot be fetched from
        the data plane or we just want to save them into the meta file.

        `None` means all fields are meta fields but it doesn't mean they
        won't be updated from the data plane.
        """
        return None

    def dump_to_dp(self) -> None:
        """Save the resource to the data plane."""

    def restore_from_dp(self) -> None:
        """Load the resource from the data plane."""

    def delete_from_dp(self) -> None:
        """Delete the resource from the data plane."""

    def update_on_dp(self) -> None:
        """Update the resource on the data plane."""


class MetaFileStorageAgentDriver(base.AbstractCapabilityDriver):
    """Meta driver. Handles models that partly are placed into the metafile.

    It's not possible to get all necessary information from the data plane
    pretty often. For instance, configuration files. There are hundreds or
    thousands of them in the system. How to know which files should be handled
    by the driver? A `meta file` may be used for this purpose. This file
    contains some meta information such path, uuid and so on but this file
    does not contain the information that can be fetched from the data plane.

    Particular models should be derived from `MetaDataPlaneModel` to work
    properly with the driver.
    """

    __model_map__ = {"model": MetaDataPlaneModel}

    def __init__(self, *args, meta_file: str, **kwargs):
        super().__init__()
        self._meta_file = meta_file
        self._storage = storage_common.JsonFileStorageSingleton(
            self._meta_file
        )

        # Check the model map is in the correct format
        for cap_models in self.__model_map__.values():
            if not issubclass(cap_models, MetaDataPlaneModel):
                raise TypeError(
                    f"Model {cap_models} is not a MetaDataPlaneModel"
                )

    def _load_from_meta(self, capability: str) -> list[MetaDataPlaneModel]:
        """Load the resources from the meta file.

        It loads only the `meta` part that cannot be fetched from the data plane
        or lightweight models.
        """

        if (
            capability not in self._storage
            # Reset storage with new format
            or "resources" not in self._storage.get(capability, {})
        ):
            self._storage[capability] = {"resources": {}}
        capstor = self._storage[capability]["resources"]

        cap_model = self.__model_map__[capability]
        return [
            cap_model.restore_from_simple_view(**r) for r in capstor.values()
        ]

    def _delete_from_meta(self, kind: str, uuid: sys_uuid.UUID) -> None:
        """Remove the resource from the meta file."""

        uuid = str(uuid)

        self._storage[kind]["resources"].pop(uuid)
        LOG.debug("Deleted meta resource %s", uuid)

    def _add_to_meta(
        self, capability: str, meta_object: MetaDataPlaneModel
    ) -> None:
        """Add the resource from the meta file."""

        view = meta_object.dump_to_simple_view()

        # Save only meta fields
        # TODO(akremenetsky): Handle an empty meta fields list.
        # Actually it means no fields are saved.
        if meta_fields := meta_object.get_meta_fields():
            for k in tuple(view.keys()):
                if k not in meta_fields:
                    view.pop(k)

        self._storage[capability]["resources"][str(meta_object.uuid)] = view
        LOG.debug("Saved meta resource: %s", view)

    def get_capabilities(self) -> list[str]:
        """Returns a list of capabilities supported by the driver."""
        return list(self.__model_map__.keys())

    def get(self, resource: models.Resource) -> models.Resource:
        """Find and return a resource by uuid and kind.

        It returns the resource from the data plane.
        """
        if resource.kind not in self.__model_map__:
            raise TypeError(f"The resource is not {self.__model_map__.keys()}")

        # Check the resource exists
        for r in self._load_from_meta(resource.kind):
            # Find the corresponding resource
            if r.uuid == resource.uuid:
                meta_resource = r
                break
        else:
            raise driver_exc.ResourceNotFound(resource=resource)

        meta_resource.restore_from_dp()
        return meta_resource.to_ua_resource(resource.kind)

    def list(self, capability: str) -> list[models.Resource]:
        """Lists all resources by capability."""
        if capability not in self.__model_map__:
            raise TypeError(f"The resource is not {self.__model_map__.keys()}")

        dp_objects = []
        for obj in self._load_from_meta(capability):
            try:
                obj.restore_from_dp()
                dp_objects.append(obj)
            except driver_exc.ResourceNotFound:
                LOG.error("Resource %s not found on the data plane", obj.uuid)

        return [obj.to_ua_resource(capability) for obj in dp_objects]

    def create(self, resource: models.Resource) -> models.Resource:
        """Creates a resource."""
        if resource.kind not in self.__model_map__:
            raise TypeError(f"The resource is not {self.__model_map__.keys()}")

        # Check the resource does not exist
        try:
            self.get(resource)
        except driver_exc.ResourceNotFound:
            # Desirable behavior, the resource should not exist
            pass
        else:
            raise driver_exc.ResourceAlreadyExists(resource=resource)

        # Restore object from the resource
        cap_model = self.__model_map__[resource.kind]

        # The object is lightweight since it is not stored in the data plane
        meta_obj = cap_model.from_ua_resource(resource)

        meta_obj.dump_to_dp()
        self._add_to_meta(resource.kind, meta_obj)
        return meta_obj.to_ua_resource(resource.kind)

    def update(self, resource: models.Resource) -> models.Resource:
        """Update the resource.

        The simplest implementation. Updating through recreating the resource.
        """
        if resource.kind not in self.__model_map__:
            raise TypeError(f"The resource is not {self.__model_map__.keys()}")

        for meta_obj in self._load_from_meta(resource.kind):
            if meta_obj.uuid == resource.uuid:
                break
        else:
            raise driver_exc.ResourceNotFound(resource=resource)

        meta_obj = self.__model_map__[resource.kind].from_ua_resource(resource)
        meta_obj.update_on_dp()

        # The simplest implementation, just recreate.
        self._delete_from_meta(resource.kind, resource.uuid)
        self._add_to_meta(resource.kind, meta_obj)

        new_resource = meta_obj.to_ua_resource(resource.kind)
        LOG.debug("Updated resource: %s", new_resource.uuid)
        return new_resource

    def delete(self, resource: models.Resource) -> None:
        """Delete the resource."""
        if resource.kind not in self.__model_map__:
            raise TypeError(f"The resource is not {self.__model_map__.keys()}")

        try:
            self.get(resource)
        except driver_exc.ResourceNotFound:
            # Nothing to do, the resource does not exist
            pass

        # Restore object from the resource
        cap_model = self.__model_map__[resource.kind]
        meta_obj = cap_model.from_ua_resource(resource)

        meta_obj.delete_from_dp()
        self._delete_from_meta(resource.kind, resource.uuid)
        LOG.debug("Deleted resource: %s", resource.uuid)

    def finalize(self) -> None:
        """Perform some finalization after finishing all operations.

        This method is called once after all other methods like list,
        create, update, delete are called. It can be used to do some
        finalization or cleanups like closing connections, files, etc.

        The driver iteration:
            start -> list -> [create | update | delete]* -> finalize
        """
        self._storage.persist()



import datetime
import inspect
import uuid as sys_uuid

from restalchemy.common import contexts
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage.sql import orm

from gcl_sdk.audit import constants
from gcl_iam import exceptions as iam_exceptions


class AuditRecord(
    models.ModelWithTimestamp,
    models.ModelWithUUID,
    orm.SQLStorableMixin,
):
    __tablename__ = "gcl_sdk_audit_logs"

    object_uuid = properties.property(
        types.UUID(),
        required=True,
        read_only=True,
    )
    object_type = properties.property(
        types.String(max_length=64),
        required=True,
        read_only=True,
    )
    user_uuid = properties.property(
        types.AllowNone(types.UUID()),
        default=None,
        read_only=True,
    )
    action = properties.property(
        types.String(max_length=64),
        required=True,
        read_only=True,
    )


class AuditLogSQLStorableMixin(orm.SQLStorableMixin):
    def insert(self, session=None, action: str = None, object_type=None):
        with self._get_engine().session_manager(session=session) as s:
            super().insert(session=s)
            self._write_audit_log(action, object_type, session=s)

    def update(
        self, session=None, force=False, action: str = None, object_type=None
    ):
        if force or self.is_dirty():
            with self._get_engine().session_manager(session=session) as s:
                super().update(session=s, force=force)
                self._write_audit_log(action, object_type, session=s)

    def delete(self, session=None, action: str = None, object_type=None):
        with self._get_engine().session_manager(session=session) as s:
            super().delete(session=s)
            self._write_audit_log(action, object_type, session=s)

    def _write_audit_log(
        self, action: str = None, object_type=None, session=None
    ):
        if action is None:
            action = inspect.stack()[1].function
            action = getattr(constants.Action, action, action)
        if object_type is None:
            object_type = self.get_table().name
        try:
            iam_context = contexts.get_context().iam_context
            user_uuid = getattr(
                getattr(iam_context, "token_info", None), "user_uuid", None
            )
        except (
            contexts.ContextIsNotExistsInStorage,
            iam_exceptions.NoIamSessionStored,
        ):
            user_uuid = None
        AuditRecord(
            object_uuid=getattr(
                self,
                "uuid",
                sys_uuid.UUID("00000000-0000-0000-0000-000000000000"),
            ),
            object_type=object_type,
            user_uuid=user_uuid,
            action=action,
        ).insert(session)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import uuid as sys_uuid
from unittest.mock import MagicMock

import pytest

from gcl_sdk.agents.universal.drivers.direct import DirectAgentDriver
from gcl_sdk.agents.universal.drivers import exceptions as driver_exc
from gcl_sdk.agents.universal.clients.backend import exceptions as client_exc
from gcl_sdk.agents.universal.storage import exceptions as storage_exc
from gcl_sdk.agents.universal.storage import base as storage_base
from gcl_sdk.agents.universal.dm import models


def _make_resource(
    kind: str, uuid: sys_uuid.UUID | None = None, value: dict | None = None
) -> models.Resource:
    uuid = uuid or sys_uuid.uuid4()
    value = value or {"uuid": str(uuid), "a": 1, "b": 2, "status": "ACTIVE"}
    return models.Resource.from_value(
        value, kind, target_fields=frozenset(value.keys())
    )


def _driver_with_caps(
    client_mock: MagicMock, storage_mock: MagicMock, caps: list[str]
):
    class _Drv(DirectAgentDriver):
        def get_capabilities(self) -> list[str]:
            return caps

    return _Drv(client=client_mock, storage=storage_mock)


class TestDirectDriver:
    def test_start_and_finalize(self):
        client = MagicMock()
        storage = MagicMock()

        drv = _driver_with_caps(client, storage, caps=["config"])

        drv.start()
        storage.load.assert_called_once_with()

        drv.finalize()
        storage.persist.assert_called_once_with()

    def test_validate_unsupported_kind_raises(self):
        client = MagicMock()
        storage = MagicMock()
        drv = _driver_with_caps(
            client, storage, caps=["config"]
        )  # only 'config' supported

        res = _make_resource("secret")

        with pytest.raises(TypeError):
            drv.get(res)

        with pytest.raises(TypeError):
            drv.create(res)

        with pytest.raises(TypeError):
            drv.update(res)

        with pytest.raises(TypeError):
            drv.list("secret")

        with pytest.raises(TypeError):
            drv.delete(res)

    def test_get_success_when_client_and_storage_ok(self):
        client = MagicMock()
        storage = MagicMock()
        drv = _driver_with_caps(client, storage, caps=["config"])

        res = _make_resource("config")
        tf_item = storage_base.TargetFieldItem(
            kind=res.kind,
            uuid=res.uuid,
            fields=frozenset({"a", "b", "uuid", "status"}),
        )

        # Note: DirectAgentDriver.get currently
        # calls storage.get(resource.uuid, resource.kind)
        storage.get.return_value = tf_item
        client_value = {
            "uuid": str(res.uuid),
            "a": 10,
            "b": 20,
            "status": "ACTIVE",
        }
        client.get.return_value = client_value

        actual = drv.get(res)

        # Ensure storage and client were called
        assert storage.get.call_count == 1
        client.get.assert_called_once_with(res)

        assert isinstance(actual, models.Resource)
        assert actual.uuid == res.uuid
        assert actual.kind == res.kind
        assert actual.value == client_value

    def test_get_not_found_maps_exceptions(self):
        client = MagicMock()
        storage = MagicMock()
        drv = _driver_with_caps(client, storage, caps=["config"])

        res = _make_resource("config")

        # storage raises -> driver maps to ResourceNotFound
        storage.get.side_effect = storage_exc.ItemNotFound(item=None)  # type: ignore[arg-type]
        with pytest.raises(driver_exc.ResourceNotFound):
            drv.get(res)

        # client raises -> driver maps to ResourceNotFound
        storage.get.reset_mock()
        storage.get.return_value = storage_base.TargetFieldItem(
            res.kind, res.uuid, frozenset(res.value.keys())
        )
        client.get.side_effect = client_exc.ResourceNotFound(resource=res)  # type: ignore[arg-type]
        with pytest.raises(driver_exc.ResourceNotFound):
            drv.get(res)

    def test_list_collects_intersection_of_storage_and_client(self):
        client = MagicMock()
        storage = MagicMock()
        drv = _driver_with_caps(client, storage, caps=["config"])

        # Storage has two items
        uuid1, uuid2, uuid3 = (
            sys_uuid.uuid4(),
            sys_uuid.uuid4(),
            sys_uuid.uuid4(),
        )
        s_item1 = storage_base.TargetFieldItem(
            "config", uuid1, frozenset({"a", "b", "uuid"})
        )
        s_item2 = storage_base.TargetFieldItem(
            "config", uuid2, frozenset({"a", "b", "uuid"})
        )
        storage.list.return_value = [s_item1, s_item2]

        # Client returns: one dict that matches storage (uuid1),
        # one Resource that doesn't have a storage entry (uuid3),
        # and one dict for uuid2 also matches storage.
        client_list = [
            {"uuid": str(uuid1), "a": 1, "b": 2},
            models.Resource.from_value(
                {"uuid": str(uuid3), "a": 3, "b": 4}, "config", s_item2.fields
            ),
            {"uuid": str(uuid2), "a": 5, "b": 6},
        ]
        client.list.return_value = client_list

        resources = drv.list("config")

        # Only items with both client and storage
        # should be returned (uuid1, uuid2)
        uuids = {r.uuid for r in resources}
        assert uuids == {uuid1, uuid2}
        assert all(r.kind == "config" for r in resources)

        storage.list.assert_called_once_with("config")
        client.list.assert_called_once_with("config")

    def test_create_success_and_storage_create_called_before_client(self):
        client = MagicMock()
        storage = MagicMock()
        drv = _driver_with_caps(client, storage, caps=["config"])

        res = _make_resource("config")
        client_resp = {
            "uuid": str(res.uuid),
            "a": 100,
            "b": 200,
            "status": "ACTIVE",
        }
        client.create.return_value = client_resp

        actual = drv.create(res)

        # storage.create called with TargetFieldItem and force=True
        assert storage.create.call_count == 1
        args, kwargs = storage.create.call_args
        assert isinstance(args[0], storage_base.TargetFieldItem)
        assert kwargs == {"force": True}

        # client.create called with resource
        client.create.assert_called_once_with(res)

        # Check result resource corresponds to client response
        assert actual.uuid == res.uuid
        assert actual.kind == res.kind
        assert actual.value == client_resp

        # Do not assert strict cross-mock ordering; just ensure both were called

    def test_create_maps_already_exists_exception(self):
        client = MagicMock()
        storage = MagicMock()
        drv = _driver_with_caps(client, storage, caps=["config"])

        res = _make_resource("config")
        client.create.side_effect = client_exc.ResourceAlreadyExists(resource=res)  # type: ignore[arg-type]

        with pytest.raises(driver_exc.ResourceAlreadyExists):
            drv.create(res)

    def test_update_success(self):
        client = MagicMock()
        storage = MagicMock()
        drv = _driver_with_caps(client, storage, caps=["config"])

        res = _make_resource("config")
        client_resp = {
            "uuid": str(res.uuid),
            "a": 11,
            "b": 22,
            "status": "ACTIVE",
        }
        client.update.return_value = client_resp

        actual = drv.update(res)

        # storage.update called with TargetFieldItem
        assert storage.update.call_count == 1
        args, _ = storage.update.call_args
        assert isinstance(args[0], storage_base.TargetFieldItem)

        client.update.assert_called_once_with(res)
        assert actual.value == client_resp

    def test_update_maps_not_found_exception(self):
        client = MagicMock()
        storage = MagicMock()
        drv = _driver_with_caps(client, storage, caps=["config"])

        res = _make_resource("config")
        client.update.side_effect = client_exc.ResourceNotFound(resource=res)  # type: ignore[arg-type]

        with pytest.raises(driver_exc.ResourceNotFound):
            drv.update(res)

    def test_delete_calls_client_and_storage_even_if_client_not_found(self):
        client = MagicMock()
        storage = MagicMock()
        drv = _driver_with_caps(client, storage, caps=["config"])

        res = _make_resource("config")

        # Case 1: normal delete
        drv.delete(res)
        client.delete.assert_called_once_with(res)
        # Ensure storage.delete was called with force=True and matching identity
        assert storage.delete.call_count == 1
        del_args, del_kwargs = storage.delete.call_args
        assert del_kwargs == {"force": True}
        item = del_args[0]
        if isinstance(item, storage_base.TargetFieldItem):
            assert item.kind == res.kind and item.uuid == res.uuid
        else:
            assert (
                getattr(item, "kind", None) == res.kind
                and getattr(item, "uuid", None) == res.uuid
            )

        # Case 2: client raises not found, storage.delete still called
        client.delete.reset_mock()
        storage.delete.reset_mock()
        client.delete.side_effect = client_exc.ResourceNotFound(resource=res)  # type: ignore[arg-type]

        drv.delete(res)

        client.delete.assert_called_once_with(res)
        assert storage.delete.call_count == 1
        del_args, del_kwargs = storage.delete.call_args
        assert del_kwargs == {"force": True}
        item = del_args[0]
        if isinstance(item, storage_base.TargetFieldItem):
            assert item.kind == res.kind and item.uuid == res.uuid
        else:
            assert (
                getattr(item, "kind", None) == res.kind
                and getattr(item, "uuid", None) == res.uuid
            )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_config import cfg
from restalchemy.dm import filters as ra_filters

from gcl_sdk.events import clients
from gcl_sdk.events import constants as event_c
from gcl_sdk.events.dm import models
from gcl_looper.services import basic


LOG = logging.getLogger(__name__)


class EventSenderService(basic.BasicService):

    def __init__(
        self, notification_client, butch_size=100, enabled=False, **kwargs
    ):
        super().__init__(**kwargs)
        self._notification_client = notification_client
        self._butch_size = butch_size
        self._enabled = enabled

    def _fetch_new_events(self):
        return models.Event.objects.get_all(
            filters={
                "status": ra_filters.EQ(models.Event.STATUS.NEW.value),
            },
            limit=self._butch_size,
        )

    def _fetch_in_progress_events(self):
        return models.Event.objects.get_all(
            filters={
                "status": ra_filters.EQ(models.Event.STATUS.IN_PROGRESS.value),
            },
            limit=self._butch_size,
        )

    def _send_new_events(self):
        events = self._fetch_new_events()
        for event in events:
            try:
                self._notification_client.send_event(event)
                event.status = models.Event.STATUS.IN_PROGRESS.value
            except:
                LOG.exception("Can't send event by reason:")
            event.save()

    def _process_status_events(self):
        events = self._fetch_in_progress_events()
        for event in events:
            try:
                event_status = self._notification_client.get_event_status(
                    event_uuid=event.uuid
                )
                if event_status == models.Event.STATUS.ACTIVE.value:
                    event.status = event_status
                else:
                    event.status = models.Event.STATUS.IN_PROGRESS.value
            except clients.EventNotFound:
                event.status = models.Event.STATUS.NEW.value
            event.save()

    @classmethod
    def build_from_config(cls, conf=None, client_cls=None, **kwargs):
        conf = conf or cfg.CONF
        client_cls = client_cls or clients.HttpEventClient

        notification_client = client_cls.build_from_config(conf=conf)

        return cls(
            enabled=conf[event_c.DOMAIN].enabled,
            notification_client=notification_client,
        )

    def _iteration(self):
        if self._enabled:
            self._send_new_events()
            self._process_status_events()
        LOG.debug(
            "Process Events is %s", "enabled" if self._enabled else "disabled"
        )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from logging import config as logging_config
import sys

from oslo_config import cfg
import yaml


DEFAULT_CONFIG = {
    "version": 1,
    "formatters": {
        "aardvark": {
            "datefmt": "%Y-%m-%d,%H:%M:%S",
            "format": "%(asctime)15s.%(msecs)03d %(processName)s"
            " pid:%(process)d tid:%(thread)d %(levelname)s"
            " %(name)s:%(lineno)d %(message)s",
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "aardvark",
            "stream": "ext://sys.stderr",
        },
    },
    "loggers": {
        "node-manager": {},
    },
    "root": {"level": "DEBUG", "handlers": ["console"]},
}


logging_opts = [
    cfg.StrOpt(
        "config",
        default="logging.yaml",
        help="Logging subsystem configuration YAML file",
    )
]


cfg.CONF.register_cli_opts(logging_opts, "logging")


class ConfigNotFound(Exception):
    pass


def configure():
    config = cfg.CONF.logging.config
    config_file = cfg.CONF.find_file(config)

    if config_file is None:
        config_data = DEFAULT_CONFIG
    else:
        with open(config_file) as f:
            config_data = yaml.safe_load(f)

    logging_config.dictConfig(config_data)

    if config_data == DEFAULT_CONFIG:
        logging.getLogger(__name__).warning(
            "Logging configuration %s not found - using defaults", config
        )


def die(logger, message):
    logger.error(message)
    sys.exit(1)
